# MicroPython asyncio module
# MIT license; Copyright (c) 2019 Damien P. George

from .core import *

__version__ = (3, 0, 0)

_attrs = {
    "wait_for": "funcs",
    "wait_for_ms": "funcs",
    "gather": "funcs",
    "Event": "event",
    "ThreadSafeFlag": "event",
    "Lock": "lock",
    "open_connection": "stream",
    "start_server": "stream",
    "StreamReader": "stream",
    "StreamWriter": "stream",
}


# Lazy loader, effectively does:
#   global attr
#   from .mod import attr
def __getattr__(attr):
    mod = _attrs.get(attr, None)
    if mod is None:
        raise AttributeError(attr)
    value = getattr(__import__(mod, None, None, True, 1), attr)
    globals()[attr] = value
    return value



import argparse

parser = argparse.ArgumentParser(description="command line program")
parser.add_argument("a")
parser.add_argument("b")
parser.add_argument(dest="c")
args = parser.parse_args(["1", "2", "3"])
assert args.a == "1" and args.b == "2" and args.c == "3"

parser = argparse.ArgumentParser()
parser.add_argument("-a", action="store_true")
parser.add_argument("-b", default=123)
args = parser.parse_args([])
assert args.a is False and args.b == 123
args = parser.parse_args(["-a"])
assert args.a is True and args.b == 123
args = parser.parse_args(["-b", "456"])
assert args.a is False and args.b == "456"

parser = argparse.ArgumentParser()
parser.add_argument("-a", "--a-opt", action="store_true")
parser.add_argument("-b", "--b-opt", default=123)
parser.add_argument("--c-opt", default="test")
args = parser.parse_args([])
assert args.a_opt is False and args.b_opt == 123 and args.c_opt == "test"
args = parser.parse_args(["--a-opt"])
assert args.a_opt is True and args.b_opt == 123 and args.c_opt == "test"
args = parser.parse_args(["--b-opt", "456"])
assert args.a_opt is False and args.b_opt == "456" and args.c_opt == "test"
args = parser.parse_args(["--c-opt", "override"])
assert args.a_opt is False and args.b_opt == 123 and args.c_opt == "override"

parser = argparse.ArgumentParser()
parser.add_argument("files", nargs="+")
args = parser.parse_args(["a"])
assert args.files == ["a"]
args = parser.parse_args(["a", "b", "c"])
assert args.files == ["a", "b", "c"]

parser = argparse.ArgumentParser()
parser.add_argument("files1", nargs=2)
parser.add_argument("files2", nargs="*")
args = parser.parse_args(["a", "b"])
assert args.files1 == ["a", "b"] and args.files2 == []
args = parser.parse_args(["a", "b", "c"])
assert args.files1 == ["a", "b"] and args.files2 == ["c"]

parser = argparse.ArgumentParser()
parser.add_argument("a", nargs=2)
parser.add_argument("-b")
args, rest = parser.parse_known_args(["a", "b", "-b", "2"])
assert args.a == ["a", "b"] and args.b == "2"
assert rest == []
args, rest = parser.parse_known_args(["-b", "2", "a", "b", "c"])
assert args.a == ["a", "b"] and args.b == "2"
assert rest == ["c"]
args, rest = parser.parse_known_args(["a", "b", "-b", "2", "c"])
assert args.a == ["a", "b"] and args.b == "2"
assert rest == ["c"]
args, rest = parser.parse_known_args(["-b", "2", "a", "b", "-", "c"])
assert args.a == ["a", "b"] and args.b == "2"
assert rest == ["-", "c"]
args, rest = parser.parse_known_args(["a", "b", "-b", "2", "-", "x", "y"])
assert args.a == ["a", "b"] and args.b == "2"
assert rest == ["-", "x", "y"]
args, rest = parser.parse_known_args(["a", "b", "c", "-b", "2", "--x", "5", "1"])
assert args.a == ["a", "b"] and args.b == "2"
assert rest == ["c", "--x", "5", "1"]



# Test machine.UART.IRQ_TXIDLE firing after a transmission.
# Does not require any external connections.

try:
    from machine import UART

    UART.IRQ_TXIDLE
except (ImportError, AttributeError):
    print("SKIP")
    raise SystemExit

import time, sys

# Configure pins based on the target.
if "rp2" in sys.platform:
    uart_id = 0
    tx_pin = "GPIO0"
    rx_pin = "GPIO1"
elif "samd" in sys.platform and "ItsyBitsy M0" in sys.implementation._machine:
    uart_id = 0
    tx_pin = "D1"
    rx_pin = "D0"
elif "samd" in sys.platform and "ItsyBitsy M4" in sys.implementation._machine:
    uart_id = 3
    tx_pin = "D1"
    rx_pin = "D0"
elif "mimxrt" in sys.platform:
    uart_id = 1
    tx_pin = None
elif "nrf" in sys.platform:
    uart_id = 0
    tx_pin = None
else:
    print("Please add support for this test on this platform.")
    raise SystemExit


def irq(u):
    print("IRQ_TXIDLE:", u.irq().flags() == u.IRQ_TXIDLE)


text = "Hello World" * 20

# Test that the IRQ is called after the write has completed.
for bits_per_s in (2400, 9600, 115200):
    if tx_pin is None:
        uart = UART(uart_id, bits_per_s)
    else:
        uart = UART(uart_id, bits_per_s, tx=tx_pin, rx=rx_pin)

    uart.irq(irq, uart.IRQ_TXIDLE)

    # The IRQ_TXIDLE shall trigger after the message has been sent. Thus
    # the test marks a time window close to the expected of the sending
    # and the time at which the IRQ should have been fired.
    # It is just a rough estimation of 10 characters before and
    # 20 characters after the data's end.

    bits_per_char = 10  # 1(startbit) + 8(bits) + 1(stopbit) + 0(parity)
    start_time_us = (len(text) - 10) * bits_per_char * 1_000_000 // bits_per_s
    window_ms = 20 * bits_per_char * 1_000 // bits_per_s + 1

    print("write", bits_per_s)
    uart.write(text)
    time.sleep_us(start_time_us)
    print("ready")
    time.sleep_ms(window_ms)
    print("done")



import requests

r = requests.get("http://api.xively.com/")
print(r)
print(r.content)
print(r.text)
print(r.content)
print(r.json())

# It's mandatory to close response objects as soon as you finished
# working with them. On MicroPython platforms without full-fledged
# OS, not doing so may lead to resource leaks and malfunction.
r.close()



#!/usr/bin/env python3
#
# This file is part of the MicroPython project, http://micropython.org/
#
# The MIT License (MIT)
#
# Copyright (c) 2022 Andrew Leech
# Copyright (c) 2022 Jim Mussared
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function
import os
import re
import stat
import subprocess

NATIVE_ARCHS = {
    "NATIVE_ARCH_NONE": "",
    "NATIVE_ARCH_X86": "x86",
    "NATIVE_ARCH_X64": "x64",
    "NATIVE_ARCH_ARMV6": "armv6",
    "NATIVE_ARCH_ARMV6M": "armv6m",
    "NATIVE_ARCH_ARMV7M": "armv7m",
    "NATIVE_ARCH_ARMV7EM": "armv7em",
    "NATIVE_ARCH_ARMV7EMSP": "armv7emsp",
    "NATIVE_ARCH_ARMV7EMDP": "armv7emdp",
    "NATIVE_ARCH_XTENSA": "xtensa",
    "NATIVE_ARCH_XTENSAWIN": "xtensawin",
    "NATIVE_ARCH_RV32IMC": "rv32imc",
}

globals().update(NATIVE_ARCHS)

__all__ = ["version", "compile", "run", "CrossCompileError"] + list(NATIVE_ARCHS.keys())


class CrossCompileError(Exception):
    pass


_VERSION_RE = re.compile("mpy-cross emitting mpy v([0-9]+)(?:.([0-9]+))?")


def _find_mpy_cross_binary(mpy_cross):
    if mpy_cross:
        return mpy_cross
    return os.path.abspath(os.path.join(os.path.dirname(__file__), "../build/mpy-cross"))


def mpy_version(mpy_cross=None):
    """
    Get the version and sub-version of the .mpy file format generated by this version of mpy-cross.

    Returns: A tuple of `(mpy_version, mpy_sub_version)`
    Optional keyword arguments:
     - mpy_cross: Specific mpy-cross binary to use
    """
    version_info = run(["--version"], mpy_cross=mpy_cross)
    match = re.search(_VERSION_RE, version_info)
    mpy_version, mpy_sub_version = int(match.group(1)), int(match.group(2) or "0")
    return (
        mpy_version,
        mpy_sub_version,
    )


def compile(src, dest=None, src_path=None, opt=None, march=None, mpy_cross=None, extra_args=None):
    """
    Compile the specified .py file with mpy-cross.

    Returns: Standard output from mpy-cross as a string.

    Required arguments:
     - src:        The path to the .py file

    Optional keyword arguments:
     - dest:       The output .mpy file. Defaults to `src` (with .mpy extension)
     - src_path:   The path to embed in the .mpy file (defaults to `src`)
     - opt:        Optimisation level (0-3, default 0)
     - march:      One of the `NATIVE_ARCH_*` constants (defaults to NATIVE_ARCH_NONE)
     - mpy_cross:  Specific mpy-cross binary to use
     - extra_args: Additional arguments to pass to mpy-cross (e.g. `["-X", "emit=native"]`)
    """
    if not src:
        raise ValueError("src is required")
    if not os.path.exists(src):
        raise CrossCompileError("Input .py file not found: {}.".format(src))

    args = []

    if src_path:
        args += ["-s", src_path]

    if dest:
        args += ["-o", dest]

    if march:
        args += ["-march=" + march]

    if opt is not None:
        args += ["-O{}".format(opt)]

    if extra_args:
        args += extra_args

    args += [src]

    run(args, mpy_cross)


def run(args, mpy_cross=None):
    """
    Run mpy-cross with the specified command line arguments.
    Prefer to use `compile()` instead.

    Returns: Standard output from mpy-cross as a string.

    Optional keyword arguments:
     - mpy_cross: Specific mpy-cross binary to use
    """
    mpy_cross = _find_mpy_cross_binary(mpy_cross)

    if not os.path.exists(mpy_cross) and not os.path.exists(mpy_cross + ".exe"):
        raise CrossCompileError("mpy-cross binary not found at {}.".format(mpy_cross))

    try:
        st = os.stat(mpy_cross)
        os.chmod(mpy_cross, st.st_mode | stat.S_IEXEC)
    except OSError:
        pass

    try:
        return subprocess.check_output([mpy_cross] + args, stderr=subprocess.STDOUT).decode()
    except subprocess.CalledProcessError as er:
        raise CrossCompileError(er.output.decode())



from machine import SPI
# Regression test for DMA for DCache coherency bugs with cache line
# written originally for https://github.com/micropython/micropython/issues/13471

# IMPORTANT: This test requires SPI2 MISO (pin Y8 on Pyboard D) to be connected to GND

SPI_NUM = 2

spi = SPI(SPI_NUM, baudrate=5_000_000)
buf = bytearray(1024)
ok = True

for offs in range(0, len(buf)):
    v = memoryview(buf)[offs : offs + 128]
    spi.readinto(v, 0xFF)
    if not all(b == 0x00 for b in v):
        print(offs, v.hex())
        ok = False

print("Variable offset fixed length " + ("OK" if ok else "FAIL"))

# this takes around 30s to run, so skipped if already failing
if ok:
    for op_len in range(1, 66):
        wr = b"\xFF" * op_len
        for offs in range(1, len(buf) - op_len - 1):
            # Place some "sentinel" values before and after the DMA buffer
            before = offs & 0xFF
            after = (~offs) & 0xFF
            buf[offs - 1] = before
            buf[offs + op_len] = after
            v = memoryview(buf)[offs : offs + op_len]
            spi.write_readinto(wr, v)
            if (
                not all(b == 0x00 for b in v)
                or buf[offs - 1] != before
                or buf[offs + op_len] != after
            ):
                print(v.hex())
                print(hex(op_len), hex(offs), hex(buf[offs - 1]), hex(buf[offs + op_len]))
                ok = False

    print("Variable offset and lengths " + ("OK" if ok else "FAIL"))



# test ssl verify_callback

import io
import socket
import tls


def verify_callback(cert, depth):
    print("verify_callback:", type(cert), len(cert) > 100, depth)
    return 0


def verify_callback_fail(cert, depth):
    print("verify_callback_fail:", type(cert), len(cert) > 100, depth)
    return 1


def test(peer_addr):
    context = tls.SSLContext(tls.PROTOCOL_TLS_CLIENT)
    context.verify_mode = tls.CERT_OPTIONAL
    context.verify_callback = verify_callback
    s = socket.socket()
    s.connect(peer_addr)
    s = context.wrap_socket(s)
    s.close()

    context.verify_callback = verify_callback_fail
    s = socket.socket()
    s.connect(peer_addr)
    try:
        s = context.wrap_socket(s)
    except OSError as e:
        print(e.args[1])


if __name__ == "__main__":
    test(socket.getaddrinfo("micropython.org", 443)[0][-1])



# test for native generators


# simple generator with yield and return
@micropython.native
def gen1(x):
    yield x
    yield x + 1
    return x + 2


g = gen1(3)
print(next(g))
print(next(g))
try:
    next(g)
except StopIteration as e:
    print(e.args[0])


# using yield from
@micropython.native
def gen2(x):
    yield from range(x)


print(list(gen2(3)))


# catching an exception from .throw()
@micropython.native
def gen3():
    try:
        yield 1
        yield 2
    except Exception as er:
        print("caught", repr(er))
        yield 3


g = gen3()
print(next(g))
print(g.throw(ValueError(42)))


# responding to .close()
@micropython.native
def gen4():
    try:
        yield 1
    except:
        print("raising GeneratorExit")
        raise GeneratorExit


g = gen4()
print(next(g))
print(g.close())



#!/usr/bin/env python3
"""
Reproducible builds module for the NEAR Python contract compiler.
"""

import subprocess
import tomllib
from pathlib import Path
from typing import Any, Dict, List

import tomli_w

from .utils import console, is_running_in_container, run_command_with_progress


def get_git_info(contract_dir: Path) -> Dict[str, Any]:
    """
    Get Git repository information for the contract.

    Args:
        contract_dir: Path to the contract directory

    Returns:
        Dictionary with Git information or empty dict if not a Git repo
    """
    git_info: dict = {}

    try:
        # Check if in git repository
        result = subprocess.run(
            ["git", "rev-parse", "--is-inside-work-tree"],
            cwd=contract_dir,
            capture_output=True,
            text=True,
            check=False,
        )

        if result.returncode != 0:
            return git_info

        # Get remote URL
        result = subprocess.run(
            ["git", "config", "--get", "remote.origin.url"],
            cwd=contract_dir,
            capture_output=True,
            text=True,
            check=False,
        )

        if result.returncode == 0:
            git_info["repository"] = result.stdout.strip()

        # Get current commit hash
        result = subprocess.run(
            ["git", "rev-parse", "HEAD"],
            cwd=contract_dir,
            capture_output=True,
            text=True,
            check=False,
        )

        if result.returncode == 0:
            git_info["commit"] = result.stdout.strip()

        # Check if working tree is clean
        result = subprocess.run(
            ["git", "status", "--porcelain"],
            cwd=contract_dir,
            capture_output=True,
            text=True,
            check=False,
        )

        git_info["clean"] = result.stdout.strip() == ""

        return git_info
    except Exception as e:
        console.print(f"[yellow]Warning: Error getting Git info: {e}")
        return {}


def verify_git_status(contract_dir: Path) -> bool:
    """
    Verify that the Git repository is clean and has a remote.

    Args:
        contract_dir: Path to the contract directory

    Returns:
        True if the repository is clean and has a remote, False otherwise
    """
    # Skip Git checks if we're in a container - the container might have
    # received files without Git history
    if is_running_in_container():
        console.print("[yellow]Running in container, skipping Git status checks")
        return True

    git_info = get_git_info(contract_dir)

    if not git_info:
        console.print("[red]Error: Not a Git repository")
        return False

    if not git_info.get("repository"):
        console.print("[red]Error: Git repository has no remote URL")
        return False

    if not git_info.get("clean", False):
        console.print(
            "[red]Error: Git working tree is not clean. Please commit all changes."
        )
        return False

    return True


def read_reproducible_build_config(contract_dir: Path) -> Dict[str, Any]:
    """
    Read reproducible build configuration from pyproject.toml.

    Args:
        contract_dir: Path to the contract directory

    Returns:
        Dictionary with reproducible build configuration
    """
    pyproject_path = contract_dir / "pyproject.toml"

    if not pyproject_path.exists():
        console.print("[red]Error: No pyproject.toml found")
        return {}

    try:
        with open(pyproject_path, "rb") as f:
            pyproject_data = tomllib.load(f)

        # Check for reproducible build configuration
        reproducible_build = (
            pyproject_data.get("tool", {}).get("near", {}).get("reproducible_build", {})
        )

        if not reproducible_build:
            console.print(
                "[yellow]Warning: No reproducible build configuration found in pyproject.toml"
            )
            console.print(
                "[yellow]Add [tool.near.reproducible_build] section with image, image_digest, and container_build_command"
            )

        return reproducible_build
    except Exception as e:
        console.print(f"[red]Error reading pyproject.toml: {e}")
        return {}


def run_reproducible_build(
    contract_path: Path, output_path: Path, build_args: List[str]
) -> bool:
    """
    Run a reproducible build in a Docker container.

    Args:
        contract_path: Path to the contract file
        output_path: Path to the output WASM file
        build_args: Arguments for the build command

    Returns:
        True if successful, False otherwise
    """
    contract_dir = contract_path.parent

    # Check if we're already in a container
    if is_running_in_container():
        console.print(
            "[yellow]Warning: Already running in a container, skipping container build"
        )
        console.print("[cyan]Running local build instead...")

        # Add --create-venv flag to ensure dependencies are installed
        if "--create-venv" not in build_args:
            build_args.append("--create-venv")

        # Run nearc directly
        import sys

        from .cli import main

        # Prepare args for click command
        cli_args = [str(contract_path)]
        if output_path:
            cli_args.extend(["--output", str(output_path)])
        cli_args.extend(build_args)

        # Replace sys.argv and run main
        sys.argv = ["nearc"] + cli_args
        main()
        return True

    # Verify Git status
    if not verify_git_status(contract_dir):
        return False

    # Read reproducible build configuration
    config = read_reproducible_build_config(contract_dir)

    if not config:
        return False

    # Check required configuration
    required_fields = ["image", "image_digest", "container_build_command"]
    for field in required_fields:
        if field not in config:
            console.print(
                f"[red]Error: Missing '{field}' in reproducible build configuration"
            )
            return False

    # Extract configuration
    image = config["image"]
    image_digest = config["image_digest"]
    container_build_command = config["container_build_command"]

    # Construct Docker image with digest
    docker_image = f"{image}@{image_digest}"

    # Get absolute paths
    abs_contract_dir = contract_dir.resolve()
    rel_contract_path = contract_path.name
    rel_output_path = output_path.name

    # Add additional build args if provided
    build_command = container_build_command + [rel_contract_path] + build_args
    if "--output" not in " ".join(build_command) and "-o" not in " ".join(
        build_command
    ):
        build_command.extend(["-o", rel_output_path])

    # Add create-venv flag to ensure dependencies are installed in the container
    if "--create-venv" not in " ".join(build_command):
        build_command.append("--create-venv")

    # Run Docker container
    docker_cmd = [
        "docker",
        "run",
        "--rm",
        "-v",
        f"{abs_contract_dir}:/home/near/code",
        "-w",
        "/home/near/code",
        "--entrypoint",
        "/bin/sh",
        docker_image,
        "-c",
        " ".join(build_command),
    ]

    console.print(
        f"[cyan]Running reproducible build in Docker container: {docker_image}"
    )
    console.print(f"[cyan]Build command: {' '.join(build_command)}")

    if not run_command_with_progress(docker_cmd, cwd=contract_dir):
        console.print("[red]Failed to run reproducible build in Docker container")
        return False

    if not output_path.exists():
        console.print(f"[red]Error: Output file {output_path} was not created")
        return False

    # Show success message with file size
    size_kb = output_path.stat().st_size / 1024
    console.print(
        f"[bold green]Successfully built reproducible contract:[/] [cyan]{output_path}[/] [yellow]({size_kb:.1f} KB)[/]"
    )

    return True


def init_reproducible_build_config(contract_dir: Path) -> bool:
    """
    Initialize reproducible build configuration in pyproject.toml.

    Args:
        contract_dir: Path to the contract directory

    Returns:
        True if successful, False otherwise
    """
    pyproject_path = contract_dir / "pyproject.toml"

    if not pyproject_path.exists():
        # Create a new pyproject.toml file
        with open(pyproject_path, "w") as f:
            f.write("[project]\n")
            f.write('name = "near-contract"\n')
            f.write('version = "0.1.0"\n')
            f.write('requires-python = ">=3.11"\n\n')

    try:
        with open(pyproject_path, "rb") as f:
            pyproject_data = tomllib.load(f)

        # Ensure tool.near section exists
        if "tool" not in pyproject_data:
            pyproject_data["tool"] = {}
        if "near" not in pyproject_data["tool"]:
            pyproject_data["tool"]["near"] = {}

        # Add reproducible build configuration
        pyproject_data["tool"]["near"]["reproducible_build"] = {
            "image": "ghcr.io/r-near/nearc:main",
            "image_digest": "sha256:REPLACE_THIS_WITH_ACTUAL_DIGEST",
            "container_build_command": ["nearc"],
        }

        # Write updated pyproject.toml
        with open(pyproject_path, "wb") as f:
            tomli_w.dump(pyproject_data, f)

        console.print(
            "[green]Initialized reproducible build configuration in pyproject.toml"
        )
        console.print(
            "[yellow]Please update the image_digest with the actual digest of the Docker image"
        )

        return True
    except Exception as e:
        console.print(f"[red]Error updating pyproject.toml: {e}")
        return False



# test subclassing a native exception


class MyExc(Exception):
    pass


e = MyExc(100, "Some error")
print(e)
print(repr(e))
print(e.args)

try:
    raise MyExc("Some error", 1)
except MyExc as e:
    print("Caught exception:", repr(e))

try:
    raise MyExc("Some error2", 2)
except Exception as e:
    print("Caught exception:", repr(e))

try:
    raise MyExc("Some error2")
except:
    print("Caught user exception")


class MyStopIteration(StopIteration):
    pass


print(MyStopIteration().value)
print(MyStopIteration(1).value)


class Iter:
    def __iter__(self):
        return self

    def __next__(self):
        # This exception will stop the "yield from", with a value of 3
        raise MyStopIteration(3)


def gen():
    print((yield from Iter()))
    return 4


try:
    next(gen())
except StopIteration as er:
    print(er.args)


class MyOSError(OSError):
    pass


print(MyOSError().errno)
print(MyOSError(1, "msg").errno)



def f():
    return 4
def g(_):
    return 5
def h():
    return 6

print(f'no interpolation')
print(f"no interpolation")
print(f"""no interpolation""")

x, y = 1, 2
print(f'{x}')
print(f'{x:08x}')
print(f'a {x} b {y} c')
print(f'a {x:08x} b {y} c')

print(f'a {"hello"} b')
print(f'a {f() + g("foo") + h()} b')

def foo(a, b):
    return f'{x}{y}{a}{b}'
print(foo(7, 8))

# ':' character within {...} that should not be interpreted as format specifiers.
print(f"a{[0,1,2][0:2]}")
print(f"a{[0,15,2][0:2][-1]:04x}")

# Nested '{' and '}' characters.
print(f"a{ {0,1,2}}")

# PEP-0498 specifies that handling of double braces '{{' or '}}' should
# behave like str.format.
print(f'{{}}')
print(f'{{{4*10}}}', '{40}')

# A single closing brace, unlike str.format should raise a syntax error.
# MicroPython instead raises ValueError at runtime from the substitution.
try:
    eval("f'{{}'")
except (ValueError, SyntaxError):
    # MicroPython incorrectly raises ValueError here.
    print('SyntaxError')

# Allow literal tuples
print(f"a {1,} b")
print(f"a {x,y,} b")
print(f"a {x,1} b")

# f-strings with conversion specifiers (only support !r and !s).
a = "123"
print(f"{a!r}")
print(f"{a!s}")
try:
    eval('print(f"{a!x}")')
except (ValueError, SyntaxError):
    # CPython detects this at compile time, MicroPython fails with ValueError
    # when the str.format is executed.
    print("ValueError")

# Mixing conversion specifiers with formatting.
print(f"{a!r:8s}")
print(f"{a!s:8s}")

# Still allow ! in expressions.
print(f"{'1' if a != '456' else '0'!r:8s}")
print(f"{'1' if a != '456' else '0'!s:8s}")

# Concatenation of adjacent f-strings.
print(f"" f"")
print(f"a" f"b")
print(f"{x}" f"{y}")
print(
    f"a{x}b---------------------------------"
    f"cd---------------------------------"
    f"e{y}f---------------------------------"
)

# Raw f-strings.
print(rf"\r\a\w {'f'} \s\t\r\i\n\g")
print(fr"\r{x}")



#!/usr/bin/env python
# MIT license; Copyright (c) 2024 Angus Gratton
#
# This is a utility script for MicroPython maintainers, similar to tools/metrics.py
# but particular to this port. It's for measuring the impact of an ESP-IDF update or
# config change at a high level.
#
# Specifically, it builds the esp32 MicroPython port for a collection of boards
# and outputs a Markdown table of binary sizes, static IRAM size, and static
# DRAM size (the latter generally inversely correlates to free heap at runtime.)
#
# To use:
#
# 1) Need to not be in an ESP-IDF venv already (i.e. don't source export.sh),
#    but IDF_PATH has to be set.
#
# 2) Choose the versions you want to test and the board/variant pairs by
#    editing the tuples below.
#
# 3) The IDF install script sometimes fails if it has to downgrade a package
#    within a minor version. The "nuclear option" is to delete all the install
#    environments and have this script recreate them as it runs:
#    rm -rf ~/.espressif/python_env/*
#
# 4) Run this script from the ports/esp32 directory, i.e.:
#    ./tools/metrics_esp32.py
#
# 5) If all goes well, it will run for a while and then print a Markdown
#    formatted table of binary sizes, sorted by board+variant.
#
# Note that for ESP32-S3 and C3, IRAM and DRAM are exchangeable so the IRAM size
# column of the table is really D/IRAM.
import os
import re
import sys
import subprocess
from dataclasses import dataclass

IDF_VERS = ("v5.2.2",)

BUILDS = (
    ("ESP32_GENERIC", ""),
    ("ESP32_GENERIC", "D2WD"),
    ("ESP32_GENERIC", "SPIRAM"),
    ("ESP32_GENERIC_S3", ""),
    ("ESP32_GENERIC_S3", "SPIRAM_OCT"),
)


@dataclass
class BuildSizes:
    idf_ver: str
    board: str
    variant: str
    bin_size: str = ""
    dram_size: str = ""
    iram_size: str = ""

    def print_summary(self, include_ver=False):
        print(f"BOARD={self.board} BOARD_VARIANT={self.variant}")
        if include_ver:
            print(f"IDF_VER {self.idf_ver}")
        print(f"Binary size: {self.bin_size}")
        print(f"IRAM size: {self.iram_size}")
        print(f"DRAM size: {self.dram_size}")

    def print_table_heading():
        print(
            "| BOARD | BOARD_VARIANT | IDF Version | Binary Size | Static IRAM Size | Static DRAM Size |"
        )
        print(
            "|-------|---------------|-------------|-------------|------------------|------------------|"
        )

    def print_table_row(self, print_board):
        print(
            "| "
            + " | ".join(
                (
                    self.board if print_board else "",
                    self.variant if print_board else "",
                    self.idf_ver,
                    self.bin_size,
                    self.iram_size,
                    self.dram_size,
                )
            )
            + " |"
        )

    def __lt__(self, other):
        """sort by board, then variant, then IDF version to get an easy
        to compare table"""
        return (self.board, self.variant, self.idf_ver) < (
            other.board,
            other.variant,
            other.idf_ver,
        )

    def build_dir(self):
        if self.variant:
            return f"build-{self.board}_{self.variant}"
        else:
            return f"build-{self.board}"

    def run_make(self, target):
        env = dict(os.environ)
        env["BOARD"] = self.board
        env["BOARD_VARIANT"] = self.variant

        try:
            # IDF version changes as we go, so re-export the environment each time
            cmd = f"source $IDF_PATH/export.sh; make {target}"
            return subprocess.check_output(
                cmd, shell=True, env=env, stderr=subprocess.STDOUT
            ).decode()
        except subprocess.CalledProcessError as e:
            err_file = f"{self.build_dir()}/make-{target}-failed-{self.idf_ver}.log"
            print(f"'make {target}' failed, writing to log to {err_file}", file=sys.stderr)
            with open(err_file, "w") as f:
                f.write(e.output.decode())
            raise

    def make_size(self):
        try:
            size_out = self.run_make("size")
            # "Used static DRAM:" or "Used stat D/IRAM:"
            RE_DRAM = r"Used stat(?:ic)? D.*: *(\d+) bytes"
            RE_IRAM = r"Used static IRAM: *(\d+) bytes"
            RE_BIN = r"Total image size: *(\d+) bytes"
            self.dram_size = re.search(RE_DRAM, size_out).group(1)
            self.iram_size = re.search(RE_IRAM, size_out).group(1)
            self.bin_size = re.search(RE_BIN, size_out).group(1)
        except subprocess.CalledProcessError:
            self.bin_size = "build failed"


def main(do_clean):
    if "IDF_PATH" not in os.environ:
        raise RuntimeError("IDF_PATH must be set")

    sizes = []
    for idf_ver in IDF_VERS:
        switch_ver(idf_ver)
        for board, variant in BUILDS:
            print(f"Building '{board}'/'{variant}'...", file=sys.stderr)
            result = BuildSizes(idf_ver, board, variant)
            result.run_make("clean")
            result.make_size()
            result.print_summary()
            sizes.append(result)

    # print everything again as a table sorted by board+variant
    last_bv = ""
    BuildSizes.print_table_heading()
    for build_sizes in sorted(sizes):
        bv = (build_sizes.board, build_sizes.variant)
        build_sizes.print_table_row(last_bv != bv)
        last_bv = bv


def idf_git(*commands):
    try:
        subprocess.check_output(
            ["git"] + list(commands), cwd=os.environ["IDF_PATH"], stderr=subprocess.STDOUT
        )
    except subprocess.CalledProcessError as e:
        print(f"git {' '.join(commands)} failed:")
        print(e.output.decode())
        raise


def idf_install():
    try:
        subprocess.check_output(
            ["bash", "install.sh"], cwd=os.environ["IDF_PATH"], stderr=subprocess.STDOUT
        )
    except subprocess.CalledProcessError as e:
        print("IDF install.sh failed:")
        print(e.output.decode())
        raise


def switch_ver(idf_ver):
    print(f"Switching version to {idf_ver}...", file=sys.stderr)
    idf_git("switch", "--detach", idf_ver)
    idf_git("submodule", "update", "--init", "--recursive")
    idf_install()


if __name__ == "__main__":
    main("--no-clean" not in sys.argv)



# base modules
include("$(PORT_DIR)/boards/manifest.py")

# asyncio
include("$(MPY_DIR)/extmod/asyncio")

# drivers
require("ssd1306")

# micropython-lib: file utilities
require("upysh")

# micropython-lib: umqtt
require("umqtt.simple")
require("umqtt.robust")



import io

c = {}


def resource_stream(package, resource):
    if package not in c:
        try:
            if package:
                p = __import__(package + ".R", None, None, True)
            else:
                p = __import__("R")
            c[package] = p.R
        except ImportError:
            if package:
                p = __import__(package)
                d = p.__path__
            else:
                d = "."
            #            if d[0] != "/":
            #                import os
            #                d = os.getcwd() + "/" + d
            c[package] = d + "/"

    p = c[package]
    if isinstance(p, dict):
        return io.BytesIO(p[resource])
    return open(p + resource, "rb")



# MicroPython LoRa STM32WL55 embedded sub-ghz radio driver
# MIT license; Copyright (c) 2022 Angus Gratton
#
# This driver is essentially an embedded SX1262 with a custom internal interface block.
# Requires the stm module in MicroPython to be compiled with STM32WL5 subghz radio support.
#
# LoRa is a registered trademark or service mark of Semtech Corporation or its affiliates.
from machine import Pin, SPI
import stm
from . import sx126x
from micropython import const

_CMD_CLR_ERRORS = const(0x07)

_REG_OCP = const(0x8E7)

# Default antenna switch config is as per Nucleo WL-55 board. See UM2592 Fig 18.
# Possible to work with other antenna switch board configurations by passing
# different ant_sw_class arguments to the modem, any class that creates an object with rx/tx


class NucleoWL55RFConfig:
    def __init__(self):
        self._FE_CTRL = (Pin(x, mode=Pin.OUT) for x in ("C4", "C5", "C3"))

    def _set_fe_ctrl(self, values):
        for pin, val in zip(self._FE_CTRL, values):
            pin(val)

    def rx(self):
        self._set_fe_ctrl((1, 0, 1))

    def tx(self, hp):
        self._set_fe_ctrl((0 if hp else 1, 1, 1))

    def idle(self):
        pass


class DIO1:
    # Dummy DIO1 "Pin" wrapper class to pass to the _SX126x class
    def irq(self, handler, _):
        stm.subghz_irq(handler)


class _WL55SubGhzModem(sx126x._SX126x):
    # Don't construct this directly, construct lora.WL55SubGhzModem or lora.AsyncWL55SubGHzModem
    def __init__(
        self,
        lora_cfg=None,
        tcxo_millivolts=1700,
        ant_sw=NucleoWL55RFConfig,
    ):
        self._hp = False

        if ant_sw == NucleoWL55RFConfig:
            # To avoid the default argument being an object instance
            ant_sw = NucleoWL55RFConfig()

        super().__init__(
            # RM0453 7.2.13 says max 16MHz, but this seems more stable
            SPI("SUBGHZ", baudrate=8_000_000),
            stm.subghz_cs,
            stm.subghz_is_busy,
            DIO1(),
            False,  # dio2_rf_sw
            tcxo_millivolts,  # dio3_tcxo_millivolts
            10_000,  # dio3_tcxo_start_time_us, first time after POR is quite long
            None,  # reset
            lora_cfg,
            ant_sw,
        )

    def _clear_errors(self):
        # A weird difference between STM32WL55 and SX1262, WL55 only takes one
        # parameter byte for the Clr_Error() command compared to two on SX1262.
        # The bytes are always zero in both cases.
        #
        # (Not clear if sending two bytes will also work always/sometimes, but
        # sending one byte to SX1262 definitely does not work!
        self._cmd("BB", _CMD_CLR_ERRORS, 0x00)

    def _clear_irq(self, clear_bits=0xFFFF):
        super()._clear_irq(clear_bits)
        # SUBGHZ Radio IRQ requires manual re-enabling after interrupt
        stm.subghz_irq(self._radio_isr)

    def _tx_hp(self):
        # STM32WL5 supports both High and Low Power antenna pins depending on tx_ant setting
        return self._hp

    def _get_pa_tx_params(self, output_power, tx_ant):
        # Given an output power level in dBm and the tx_ant setting (if any),
        # return settings for SetPaConfig and SetTxParams.
        #
        # ST document RM0453 Set_PaConfig() reference and accompanying Table 35
        # show values that are an exact superset of the SX1261 and SX1262
        # available values, depending on which antenna pin is to be
        # used. Therefore, call either modem's existing _get_pa_tx_params()
        # function depending on the current tx_ant setting (default is low
        # power).

        if tx_ant is not None:
            # Note: currently HP antenna power output is less than it should be,
            # due to some (unknown) bug.
            self._hp = tx_ant == "PA_BOOST"

        # Update the OCP register to match the maximum power level
        self._reg_write(_REG_OCP, 0x38 if self._hp else 0x18)

        if self._hp:
            return sx126x._SX1262._get_pa_tx_params(self, output_power, tx_ant)
        else:
            return sx126x._SX1261._get_pa_tx_params(self, output_power, tx_ant)


# Define the actual modem classes that use the SyncModem & AsyncModem "mixin-like" classes
# to create sync and async variants.

try:
    from .sync_modem import SyncModem

    class WL55SubGhzModem(_WL55SubGhzModem, SyncModem):
        pass

except ImportError:
    pass

try:
    from .async_modem import AsyncModem

    class AsyncWL55SubGhzModem(_WL55SubGhzModem, AsyncModem):
        pass

except ImportError:
    pass



# LOLIN S2 PICO MicroPython Helper Library

from micropython import const
from machine import Pin, I2C, Signal
from s2pico_oled import OLED

# Pin Assignments

# SPI
SPI_MOSI = const(35)
SPI_MISO = const(36)
SPI_CLK = const(37)

# I2C
I2C_SDA = const(8)
I2C_SCL = const(9)

# DAC
DAC1 = const(17)
DAC2 = const(18)

# LED
LED = const(10)

# OLED
OLED_RST = const(18)

# BUTTON
BUTTON = const(0)

# Helper methods for built in sensors

led = Signal(LED, Pin.OUT, value=0, invert=True)

button = Pin(BUTTON, Pin.IN, Pin.PULL_UP)

i2c = I2C(0)
oled = OLED(i2c, Pin(OLED_RST))



#!/usr/bin/env python3

import glob
import json
import os
import sys

VALID_FEATURES = {
    # Connectivity
    "BLE",
    "CAN",
    "Ethernet",
    "LoRa",
    "USB",
    "USB-C",
    "WiFi",
    # MCU features
    "Dual-core",
    "External Flash",
    "External RAM",
    # Form factor
    "Feather",
    # Connectors / sockets
    "JST-PH",
    "JST-SH",
    "mikroBUS",
    "microSD",
    "SDCard",
    # Sensors
    "Environment Sensor",
    "IMU",
    # Other
    "Audio Codec",
    "Battery Charging",
    "Camera",
    "DAC",
    "Display",
    "Microphone",
    "PoE",
    "RGB LED",
    "Secure Element",
}


def main(repo_path, output_path):
    boards_index = []
    board_ids = set()

    for board_json in glob.glob(os.path.join(repo_path, "ports/*/boards/*/board.json")):
        # Relative path to the board directory (e.g. "ports/stm32/boards/PYBV11").
        board_dir = os.path.dirname(board_json)
        # Relative path to the port (e.g. "ports/stm32")
        port_dir = os.path.dirname(os.path.dirname(board_dir))

        with open(board_json, "r") as f:
            blob = json.load(f)

            features = set(blob.get("features", []))
            if not features.issubset(VALID_FEATURES):
                print(
                    board_json,
                    "unknown features:",
                    features.difference(VALID_FEATURES),
                    file=sys.stderr,
                )
                sys.exit(1)

            # Use "id" if specified, otherwise default to board dir (e.g. "PYBV11").
            # We allow boards to override ID for the historical build names.
            blob["id"] = blob.get("id", os.path.basename(board_dir))

            # Check for duplicate board IDs.
            if blob["id"] in board_ids:
                print("Duplicate board ID: '{}'".format(blob["id"]), file=sys.stderr)
            board_ids.add(blob["id"])

            # Add in default fields.
            blob["port"] = os.path.basename(port_dir)
            blob["build"] = os.path.basename(board_dir)
            boards_index.append(blob)

        # Create the board markdown, which is the concatenation of the
        # default "board.md" file (if exists), as well as any flashing
        # instructions.
        board_markdown = os.path.join(board_dir, "board.md")
        with open(os.path.join(output_path, blob["id"] + ".md"), "w") as f:
            if os.path.exists(board_markdown):
                with open(board_markdown, "r") as fin:
                    f.write(fin.read())

            if blob["deploy"]:
                f.write("\n\n## Installation instructions\n")
            for deploy in blob["deploy"]:
                with open(os.path.join(board_dir, deploy), "r") as fin:
                    f.write(fin.read())

    # Write the full index for the website to load.
    with open(os.path.join(output_path, "index.json"), "w") as f:
        json.dump(boards_index, f, indent=4, sort_keys=True)
        f.write("\n")


if __name__ == "__main__":
    main(sys.argv[1], sys.argv[2])



# Test fairness of scheduler

try:
    import asyncio
except ImportError:
    print("SKIP")
    raise SystemExit


async def task(id, t):
    print("task start", id)
    while True:
        if t > 0:
            print("task work", id)
        await asyncio.sleep(t)


async def main():
    t1 = asyncio.create_task(task(1, -0.01))
    t2 = asyncio.create_task(task(2, 0.1))
    t3 = asyncio.create_task(task(3, 0.18))
    t4 = asyncio.create_task(task(4, -100))
    await asyncio.sleep(0.45)  # t2 prints 5 times, t3 prints 3 times
    t1.cancel()
    t2.cancel()
    t3.cancel()
    t4.cancel()
    print("finish")


asyncio.run(main())



import sys

print((10).to_bytes(1, "little"))
print((111111).to_bytes(4, "little"))
print((100).to_bytes(10, "little"))
print(int.from_bytes(b"\x00\x01\0\0\0\0\0\0", "little"))
print(int.from_bytes(b"\x01\0\0\0\0\0\0\0", "little"))
print(int.from_bytes(b"\x00\x01\0\0\0\0\0\0", "little"))

# check that extra zero bytes don't change the internal int value
print(int.from_bytes(bytes(20), "little") == 0)
print(int.from_bytes(b"\x01" + bytes(20), "little") == 1)

# big-endian conversion
print((10).to_bytes(1, "big"))
print((100).to_bytes(10, "big"))
print(int.from_bytes(b"\0\0\0\0\0\0\0\0\0\x01", "big"))
print(int.from_bytes(b"\x01\0", "big"))

# negative number of bytes should raise an error
try:
    (1).to_bytes(-1, "little")
except ValueError:
    print("ValueError")

# zero byte destination should also raise an error
try:
    (1).to_bytes(0, "little")
except OverflowError:
    print("OverflowError")

# except for converting 0 to a zero-length byte array
print((0).to_bytes(0, "big"))

# byte length can fit the integer directly
print((0xFF).to_bytes(1, "little"))
print((0xFF).to_bytes(1, "big"))
print((0xEFF).to_bytes(2, "little"))
print((0xEFF).to_bytes(2, "big"))
print((0xCDEFF).to_bytes(3, "little"))
print((0xCDEFF).to_bytes(3, "big"))

# OverFlowError if not big enough

try:
    (0x123).to_bytes(1, "big")
except OverflowError:
    print("OverflowError")

try:
    (0x12345).to_bytes(2, "big")
except OverflowError:
    print("OverflowError")

try:
    (0x1234567).to_bytes(3, "big")
except OverflowError:
    print("OverflowError")


# negative representations

# MicroPython int.to_bytes() behaves as if signed=True for negative numbers
if "micropython" in repr(sys.implementation):

    def to_bytes_compat(i, l, e):
        return i.to_bytes(l, e)
else:
    # Implement MicroPython compatible behaviour for CPython
    def to_bytes_compat(i, l, e):
        return i.to_bytes(l, e, signed=i < 0)


print(to_bytes_compat(-1, 1, "little"))
print(to_bytes_compat(-1, 3, "little"))
print(to_bytes_compat(-1, 1, "big"))
print(to_bytes_compat(-1, 3, "big"))
print(to_bytes_compat(-128, 1, "big"))
print(to_bytes_compat(-32768, 2, "big"))
print(to_bytes_compat(-(1 << 23), 3, "big"))

try:
    print(to_bytes_compat(-129, 1, "big"))
except OverflowError:
    print("OverflowError")

try:
    print(to_bytes_compat(-32769, 2, "big"))
except OverflowError:
    print("OverflowError")

try:
    print(to_bytes_compat(-(1 << 23) - 1, 2, "big"))
except OverflowError:
    print("OverflowError")



# This module should be imported from REPL, not run from command line.
import binascii
import hashlib
from micropython import const
import network
import os
import socket
import sys
import websocket
import _webrepl

listen_s = None
client_s = None

DEBUG = 0

_DEFAULT_STATIC_HOST = const("https://micropython.org/webrepl/")
static_host = _DEFAULT_STATIC_HOST


def server_handshake(cl):
    req = cl.makefile("rwb", 0)
    # Skip HTTP GET line.
    l = req.readline()
    if DEBUG:
        sys.stdout.write(repr(l))

    webkey = None
    upgrade = False
    websocket = False

    while True:
        l = req.readline()
        if not l:
            # EOF in headers.
            return False
        if l == b"\r\n":
            break
        if DEBUG:
            sys.stdout.write(l)
        h, v = [x.strip() for x in l.split(b":", 1)]
        if DEBUG:
            print((h, v))
        if h == b"Sec-WebSocket-Key":
            webkey = v
        elif h == b"Connection" and b"Upgrade" in v:
            upgrade = True
        elif h == b"Upgrade" and v == b"websocket":
            websocket = True

    if not (upgrade and websocket and webkey):
        return False

    if DEBUG:
        print("Sec-WebSocket-Key:", webkey, len(webkey))

    d = hashlib.sha1(webkey)
    d.update(b"258EAFA5-E914-47DA-95CA-C5AB0DC85B11")
    respkey = d.digest()
    respkey = binascii.b2a_base64(respkey)[:-1]
    if DEBUG:
        print("respkey:", respkey)

    cl.send(
        b"""\
HTTP/1.1 101 Switching Protocols\r
Upgrade: websocket\r
Connection: Upgrade\r
Sec-WebSocket-Accept: """
    )
    cl.send(respkey)
    cl.send("\r\n\r\n")

    return True


def send_html(cl):
    cl.send(
        b"""\
HTTP/1.0 200 OK\r
\r
<base href=\""""
    )
    cl.send(static_host)
    cl.send(
        b"""\"></base>\r
<script src="webrepl_content.js"></script>\r
"""
    )
    cl.close()


def setup_conn(port, accept_handler):
    global listen_s
    listen_s = socket.socket()
    listen_s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    ai = socket.getaddrinfo("0.0.0.0", port)
    addr = ai[0][4]

    listen_s.bind(addr)
    listen_s.listen(1)
    if accept_handler:
        listen_s.setsockopt(socket.SOL_SOCKET, 20, accept_handler)
    for i in (network.WLAN.IF_AP, network.WLAN.IF_STA):
        iface = network.WLAN(i)
        if iface.active():
            print("WebREPL server started on http://%s:%d/" % (iface.ifconfig()[0], port))
    return listen_s


def accept_conn(listen_sock):
    global client_s
    cl, remote_addr = listen_sock.accept()

    if not server_handshake(cl):
        send_html(cl)
        return False

    prev = os.dupterm(None)
    os.dupterm(prev)
    if prev:
        print("\nConcurrent WebREPL connection from", remote_addr, "rejected")
        cl.close()
        return False
    print("\nWebREPL connection from:", remote_addr)
    client_s = cl

    ws = websocket.websocket(cl, True)
    ws = _webrepl._webrepl(ws)
    cl.setblocking(False)
    # notify REPL on socket incoming data (ESP32/ESP8266-only)
    if hasattr(os, "dupterm_notify"):
        cl.setsockopt(socket.SOL_SOCKET, 20, os.dupterm_notify)
    os.dupterm(ws)

    return True


def stop():
    global listen_s, client_s
    os.dupterm(None)
    if client_s:
        client_s.close()
    if listen_s:
        listen_s.close()


def start(port=8266, password=None, accept_handler=accept_conn):
    global static_host
    stop()
    webrepl_pass = password
    if webrepl_pass is None:
        try:
            import webrepl_cfg

            webrepl_pass = webrepl_cfg.PASS
            if hasattr(webrepl_cfg, "BASE"):
                static_host = webrepl_cfg.BASE
        except:
            print("WebREPL is not configured, run 'import webrepl_setup'")

    _webrepl.password(webrepl_pass)
    s = setup_conn(port, accept_handler)

    if accept_handler is None:
        print("Starting webrepl in foreground mode")
        # Run accept_conn to serve HTML until we get a websocket connection.
        while not accept_conn(s):
            pass
    elif password is None:
        print("Started webrepl in normal mode")
    else:
        print("Started webrepl in manual override mode")


def start_foreground(port=8266, password=None):
    start(port, password, None)



#!/usr/bin/env python3

import re, subprocess, sys


DEBUG = False
DRY_RUN = False
NUM_KEEP_PER_BOARD = 4


def main():
    ssh_machine = sys.argv[1]
    ssh_firmware_dir = sys.argv[2]

    # SSH to get list of existing files.
    p = subprocess.run(
        ["ssh", ssh_machine, "find", ssh_firmware_dir, "-name", "\\*-preview.\\*"],
        capture_output=True,
    )
    if p.returncode != 0:
        print(p.stderr)
        return
    all_files = p.stdout.split(b"\n")

    # Parse all files to organise into boards/date/version.
    boards = {}
    for file in all_files:
        m = re.match(
            rb"([a-z/.]+)/([A-Za-z0-9_-]+)-(20[0-9]{6})-(v[0-9.]+)-preview.([0-9]+).g[0-9a-f]+.",
            file,
        )
        if not m:
            continue
        dir, board, date, version, ncommits = m.groups()
        if board not in boards:
            boards[board] = {}
        if (date, version, ncommits) not in boards[board]:
            boards[board][(date, version, ncommits)] = []
        boards[board][(date, version, ncommits)].append(file)

    # Collect files to remove based on date and version.
    remove = []
    for board in boards.values():
        filelist = [
            (date, version, ncommits, files) for (date, version, ncommits), files in board.items()
        ]
        filelist.sort(reverse=True)
        keep = []
        for date, version, ncommits, files in filelist:
            if keep and (version, ncommits) == keep[-1]:
                remove.extend(files)
            elif len(keep) >= NUM_KEEP_PER_BOARD:
                remove.extend(files)
            else:
                keep.append((version, ncommits))

    if DEBUG:
        all_files.sort(reverse=True)
        for file in all_files:
            print(file, file in remove)
        print(len(remove), "/", len(all_files))

    # Do removal of files.
    for file in remove:
        file = str(file, "ascii")
        print("remove:", file)
        if not DRY_RUN:
            p = subprocess.run(["ssh", ssh_machine, "/bin/rm", file], capture_output=True)
            if p.returncode != 0:
                print(p.stderr)


if __name__ == "__main__":
    main()



def partial(func, *args, **kwargs):
    def _partial(*more_args, **more_kwargs):
        kw = kwargs.copy()
        kw.update(more_kwargs)
        return func(*(args + more_args), **kw)

    return _partial


def update_wrapper(wrapper, wrapped, assigned=None, updated=None):
    # Dummy impl
    return wrapper


def wraps(wrapped, assigned=None, updated=None):
    # Dummy impl
    return lambda x: x


def lru_cache(*args, **kwargs):
    def lru_cache_impl(func):
        return func
    if len(args) == 1 and callable(args[0]):
        return lru_cache_impl(args[0])
    else:
        return lru_cache_impl


def reduce(function, iterable, initializer=None):
    it = iter(iterable)
    if initializer is None:
        value = next(it)
    else:
        value = initializer
    for element in it:
        value = function(value, element)
    return value



def cast(type, val):
    return val


def get_origin(type):
    return None


def get_args(type):
    return ()


def no_type_check(func):
    def wrapper(*args, **kwargs):
        func(*args, **kwargs)
    return wrapper


def overload(func):
    return None


class _AnyCall:
    def __init__(*args, **kwargs):
        pass

    def __call__(*args, **kwargs):
        pass

    def __getitem__(self, arg):
        return _anyCall

_anyCall = _AnyCall()


class _SubscriptableType:
    def __getitem__(self, arg):
        return _anyCall


_Subscriptable = _SubscriptableType()


def TypeVar(name, *types, bound: Any | None = None, covariant = False, contravariant = False, infer_variance = False):
    return None


def NewType(name, type):
    return type


class Any:
    pass


class BinaryIO:
    pass


class ClassVar:
    pass


class Final:
    pass


class Hashable:
    pass


class IO:
    pass


class NoReturn:
    pass


class Sized:
    pass


class SupportsInt:
    pass


class SupportsFloat:
    pass


class SupportsComplex:
    pass


class SupportsBytes:
    pass


class SupportsIndex:
    pass


class SupportsAbs:
    pass


class SupportsRound:
    pass


class TextIO:
    pass


AnyStr = str
Text = str
Pattern = str
Match = str
TypedDict = dict

AbstractSet = _Subscriptable
AsyncContextManager = _Subscriptable
AsyncGenerator = _Subscriptable
AsyncIterable = _Subscriptable
AsyncIterator = _Subscriptable
Awaitable = _Subscriptable
Callable = _Subscriptable
ChainMap = _Subscriptable
Collection = _Subscriptable
Container = _Subscriptable
ContextManager = _Subscriptable
Coroutine = _Subscriptable
Counter = _Subscriptable
DefaultDict = _Subscriptable
Deque = _Subscriptable
Dict = _Subscriptable
FrozenSet = _Subscriptable
Generator = _Subscriptable
Generic = _Subscriptable
Iterable = _Subscriptable
Iterator = _Subscriptable
List = _Subscriptable
Literal = _Subscriptable
Mapping = _Subscriptable
MutableMapping = _Subscriptable
MutableSequence = _Subscriptable
MutableSet = _Subscriptable
NamedTuple = _Subscriptable
Optional = _Subscriptable
OrderedDict = _Subscriptable
Sequence = _Subscriptable
Set = _Subscriptable
Tuple = _Subscriptable
Type = _Subscriptable
Union = _Subscriptable

TYPE_CHECKING = False



# 1-Wire driver for MicroPython
# MIT license; Copyright (c) 2016 Damien P. George

import _onewire as _ow


class OneWireError(Exception):
    pass


class OneWire:
    SEARCH_ROM = 0xF0
    MATCH_ROM = 0x55
    SKIP_ROM = 0xCC

    def __init__(self, pin):
        self.pin = pin
        self.pin.init(pin.OPEN_DRAIN, pin.PULL_UP)

    def reset(self, required=False):
        reset = _ow.reset(self.pin)
        if required and not reset:
            raise OneWireError
        return reset

    def readbit(self):
        return _ow.readbit(self.pin)

    def readbyte(self):
        return _ow.readbyte(self.pin)

    def readinto(self, buf):
        for i in range(len(buf)):
            buf[i] = _ow.readbyte(self.pin)

    def writebit(self, value):
        return _ow.writebit(self.pin, value)

    def writebyte(self, value):
        return _ow.writebyte(self.pin, value)

    def write(self, buf):
        for b in buf:
            _ow.writebyte(self.pin, b)

    def select_rom(self, rom):
        self.reset()
        self.writebyte(self.MATCH_ROM)
        self.write(rom)

    def scan(self):
        devices = []
        diff = 65
        rom = False
        for i in range(0xFF):
            rom, diff = self._search_rom(rom, diff)
            if rom:
                devices += [rom]
            if diff == 0:
                break
        return devices

    def _search_rom(self, l_rom, diff):
        if not self.reset():
            return None, 0
        self.writebyte(self.SEARCH_ROM)
        if not l_rom:
            l_rom = bytearray(8)
        rom = bytearray(8)
        next_diff = 0
        i = 64
        for byte in range(8):
            r_b = 0
            for bit in range(8):
                b = self.readbit()
                if self.readbit():
                    if b:  # there are no devices or there is an error on the bus
                        return None, 0
                else:
                    if not b:  # collision, two devices with different bit meaning
                        if diff > i or ((l_rom[byte] & (1 << bit)) and diff != i):
                            b = 1
                            next_diff = i
                self.writebit(b)
                if b:
                    r_b |= 1 << bit
                i -= 1
            rom[byte] = r_b
        return rom, next_diff

    def crc8(self, data):
        return _ow.crc8(data)



# test round() with large integer values and second arg

# rounding integers is an optional feature so test for it
try:
    round(1, -1)
except NotImplementedError:
    print('SKIP')
    raise SystemExit

i = 2**70

tests = [
    (i, 0), (i, -1), (i, -10), (i, 1),
    (-i, 0), (-i, -1), (-i, -10), (-i, 1),
]
for t in tests:
    print(round(*t))



# MicroPython aioble module
# MIT license; Copyright (c) 2021 Jim Mussared

from micropython import const

import asyncio
import binascii

from .core import ble, register_irq_handler, log_error


_IRQ_MTU_EXCHANGED = const(21)


# Raised by `with device.timeout()`.
class DeviceDisconnectedError(Exception):
    pass


def _device_irq(event, data):
    if event == _IRQ_MTU_EXCHANGED:
        conn_handle, mtu = data
        if device := DeviceConnection._connected.get(conn_handle, None):
            device.mtu = mtu
            if device._mtu_event:
                device._mtu_event.set()


register_irq_handler(_device_irq, None)


# Context manager to allow an operation to be cancelled by timeout or device
# disconnection. Don't use this directly -- use `with connection.timeout(ms):`
# instead.
class DeviceTimeout:
    def __init__(self, connection, timeout_ms):
        self._connection = connection
        self._timeout_ms = timeout_ms

        # We allow either (or both) connection and timeout_ms to be None. This
        # allows this to be used either as a just-disconnect, just-timeout, or
        # no-op.

        # This task is active while the operation is in progress. It sleeps
        # until the timeout, and then cancels the working task. If the working
        # task completes, __exit__ will cancel the sleep.
        self._timeout_task = None

        # This is the task waiting for the actual operation to complete.
        # Usually this is waiting on an event that will be set() by an IRQ
        # handler.
        self._task = asyncio.current_task()

        # Tell the connection that if it disconnects, it should cancel this
        # operation (by cancelling self._task).
        if connection:
            connection._timeouts.append(self)

    async def _timeout_sleep(self):
        try:
            await asyncio.sleep_ms(self._timeout_ms)
        except asyncio.CancelledError:
            # The operation completed successfully and this timeout task was
            # cancelled by __exit__.
            return

        # The sleep completed, so we should trigger the timeout. Set
        # self._timeout_task to None so that we can tell the difference
        # between a disconnect and a timeout in __exit__.
        self._timeout_task = None
        self._task.cancel()

    def __enter__(self):
        if self._timeout_ms:
            # Schedule the timeout waiter.
            self._timeout_task = asyncio.create_task(self._timeout_sleep())

    def __exit__(self, exc_type, exc_val, exc_traceback):
        # One of five things happened:
        # 1 - The operation completed successfully.
        # 2 - The operation timed out.
        # 3 - The device disconnected.
        # 4 - The operation failed for a different exception.
        # 5 - The task was cancelled by something else.

        # Don't need the connection to tell us about disconnection anymore.
        if self._connection:
            self._connection._timeouts.remove(self)

        try:
            if exc_type == asyncio.CancelledError:
                # Case 2, we started a timeout and it's completed.
                if self._timeout_ms and self._timeout_task is None:
                    raise asyncio.TimeoutError

                # Case 3, we have a disconnected device.
                if self._connection and self._connection._conn_handle is None:
                    raise DeviceDisconnectedError

                # Case 5, something else cancelled us.
                # Allow the cancellation to propagate.
                return

            # Case 1 & 4. Either way, just stop the timeout task and let the
            # exception (if case 4) propagate.
        finally:
            # In all cases, if the timeout is still running, cancel it.
            if self._timeout_task:
                self._timeout_task.cancel()


class Device:
    def __init__(self, addr_type, addr):
        # Public properties
        self.addr_type = addr_type
        self.addr = addr if len(addr) == 6 else binascii.unhexlify(addr.replace(":", ""))
        self._connection = None

    def __eq__(self, rhs):
        return self.addr_type == rhs.addr_type and self.addr == rhs.addr

    def __hash__(self):
        return hash((self.addr_type, self.addr))

    def __str__(self):
        return "Device({}, {}{})".format(
            "ADDR_PUBLIC" if self.addr_type == 0 else "ADDR_RANDOM",
            self.addr_hex(),
            ", CONNECTED" if self._connection else "",
        )

    def addr_hex(self):
        return binascii.hexlify(self.addr, ":").decode()

    async def connect(
        self,
        timeout_ms=10000,
        scan_duration_ms=None,
        min_conn_interval_us=None,
        max_conn_interval_us=None,
    ):
        if self._connection:
            return self._connection

        # Forward to implementation in central.py.
        from .central import _connect

        await _connect(
            DeviceConnection(self),
            timeout_ms,
            scan_duration_ms,
            min_conn_interval_us,
            max_conn_interval_us,
        )

        # Start the device task that will clean up after disconnection.
        self._connection._run_task()
        return self._connection


class DeviceConnection:
    # Global map of connection handle to active devices (for IRQ mapping).
    _connected = {}

    def __init__(self, device):
        self.device = device
        device._connection = self

        self.encrypted = False
        self.authenticated = False
        self.bonded = False
        self.key_size = False
        self.mtu = None

        self._conn_handle = None

        # This event is fired by the IRQ both for connection and disconnection
        # and controls the device_task.
        self._event = asyncio.ThreadSafeFlag()

        # If we're waiting for a pending MTU exchange.
        self._mtu_event = None

        # In-progress client discovery instance (e.g. services, chars,
        # descriptors) used for IRQ mapping.
        self._discover = None
        # Map of value handle to characteristic (so that IRQs with
        # conn_handle,value_handle can route to them). See
        # ClientCharacteristic._find for where this is used.
        self._characteristics = {}

        self._task = None

        # DeviceTimeout instances that are currently waiting on this device
        # and need to be notified if disconnection occurs.
        self._timeouts = []

        # Fired by the encryption update event.
        self._pair_event = None

        # Active L2CAP channel for this device.
        # TODO: Support more than one concurrent channel.
        self._l2cap_channel = None

    # While connected, this tasks waits for disconnection then cleans up.
    async def device_task(self):
        assert self._conn_handle is not None

        # Wait for the (either central or peripheral) disconnected irq.
        await self._event.wait()

        # Mark the device as disconnected.
        del DeviceConnection._connected[self._conn_handle]
        self._conn_handle = None
        self.device._connection = None

        # Cancel any in-progress operations on this device.
        for t in self._timeouts:
            t._task.cancel()

    def _run_task(self):
        self._task = asyncio.create_task(self.device_task())

    async def disconnect(self, timeout_ms=2000):
        await self.disconnected(timeout_ms, disconnect=True)

    async def disconnected(self, timeout_ms=None, disconnect=False):
        if not self.is_connected():
            return

        # The task must have been created after successful connection.
        assert self._task

        if disconnect:
            try:
                ble.gap_disconnect(self._conn_handle)
            except OSError as e:
                log_error("Disconnect", e)

        with DeviceTimeout(None, timeout_ms):
            await self._task

    # Retrieve a single service matching this uuid.
    async def service(self, uuid, timeout_ms=2000):
        result = None
        # Make sure loop runs to completion.
        async for service in self.services(uuid, timeout_ms):
            if not result and service.uuid == uuid:
                result = service
        return result

    # Search for all services (optionally by uuid).
    # Use with `async for`, e.g.
    #     async for service in device.services():
    # Note: must allow the loop to run to completion.
    # TODO: disconnection / timeout
    def services(self, uuid=None, timeout_ms=2000):
        from .client import ClientDiscover, ClientService

        return ClientDiscover(self, ClientService, self, timeout_ms, uuid)

    async def pair(self, *args, **kwargs):
        from .security import pair

        await pair(self, *args, **kwargs)

    def is_connected(self):
        return self._conn_handle is not None

    # Use with `with` to simplify disconnection and timeout handling.
    def timeout(self, timeout_ms):
        return DeviceTimeout(self, timeout_ms)

    async def exchange_mtu(self, mtu=None, timeout_ms=1000):
        if not self.is_connected():
            raise ValueError("Not connected")

        if mtu:
            ble.config(mtu=mtu)

        self._mtu_event = self._mtu_event or asyncio.ThreadSafeFlag()
        ble.gattc_exchange_mtu(self._conn_handle)
        with self.timeout(timeout_ms):
            await self._mtu_event.wait()
        return self.mtu

    # Wait for a connection on an L2CAP connection-oriented-channel.
    async def l2cap_accept(self, psm, mtu, timeout_ms=None):
        from .l2cap import accept

        return await accept(self, psm, mtu, timeout_ms)

    # Attempt to connect to a listening device.
    async def l2cap_connect(self, psm, mtu, timeout_ms=1000):
        from .l2cap import connect

        return await connect(self, psm, mtu, timeout_ms)

    # Context manager -- automatically disconnect.
    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_traceback):
        await self.disconnect()



"""
ADC test for the CC3200 based boards.
"""

from machine import ADC
import os

mch = os.uname().machine
if "LaunchPad" in mch:
    adc_pin = "GP5"
    adc_channel = 3
elif "WiPy" in mch:
    adc_pin = "GP3"
    adc_channel = 1
else:
    raise Exception("Board not supported!")

adc = ADC(0)
print(adc)
adc = ADC()
print(adc)
adc = ADC(0, bits=12)
print(adc)

apin = adc.channel(adc_channel)
print(apin)
apin = adc.channel(id=adc_channel)
print(apin)
apin = adc.channel(adc_channel, pin=adc_pin)
print(apin)
apin = adc.channel(id=adc_channel, pin=adc_pin)
print(apin)

print(apin.value() > 3000)
print(apin() > 3000)

# de-init must work
apin.deinit()
print(apin)

adc.deinit()
print(adc)
print(apin)
adc.init()
print(adc)
print(apin)
apin.init()
print(apin)
print(apin() > 3000)

# check for memory leaks...
for i in range(0, 1000):
    adc = ADC()
    apin = adc.channel(adc_channel)

# next ones should raise
try:
    adc = ADC(bits=17)
except:
    print("Exception")

try:
    adc = ADC(id=1)
except:
    print("Exception")

try:
    adc = ADC(0, 16)
except:
    print("Exception")

adc = ADC()
try:
    apin = adc.channel(4)
except:
    print("Exception")

try:
    apin = adc.channel(-1)
except:
    print("Exception")

try:
    apin = adc.channel(0, pin="GP3")
except:
    print("Exception")

apin = adc.channel(1)
apin.deinit()
try:
    apin()
except:
    print("Exception")

try:
    apin.value()
except:
    print("Exception")

adc.deinit()
try:
    apin.value()
except:
    print("Exception")

try:
    apin = adc.channel(1)
except:
    print("Exception")

# re-init must work
adc.init()
apin.init()
print(apin)
print(apin() > 3000)



# DotStar strip driver for MicroPython
#
# The MIT License (MIT)
#
# Copyright (c) 2016 Damien P. George (original Neopixel object)
# Copyright (c) 2017 Ladyada
# Copyright (c) 2017 Scott Shawcroft for Adafruit Industries
# Copyright (c) 2019 Matt Trentini (porting back to MicroPython)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

START_HEADER_SIZE = 4
LED_START = 0b11100000  # Three "1" bits, followed by 5 brightness bits

# Pixel color order constants
RGB = (0, 1, 2)
RBG = (0, 2, 1)
GRB = (1, 0, 2)
GBR = (1, 2, 0)
BRG = (2, 0, 1)
BGR = (2, 1, 0)


class DotStar:
    """
    A sequence of dotstars.

    :param SPI spi: The SPI object to write output to.
    :param int n: The number of dotstars in the chain
    :param float brightness: Brightness of the pixels between 0.0 and 1.0
    :param bool auto_write: True if the dotstars should immediately change when
        set. If False, `show` must be called explicitly.
    :param tuple pixel_order: Set the pixel order on the strip - different
         strips implement this differently. If you send red, and it looks blue
         or green on the strip, modify this! It should be one of the values above


    Example for TinyPICO:

    .. code-block:: python

        from dotstar import DotStar
        from machine import Pin, SPI

        spi = SPI(sck=Pin(12), mosi=Pin(13), miso=Pin(18)) # Configure SPI - note: miso is unused
        dotstar = DotStar(spi, 1)
        dotstar[0] = (128, 0, 0) # Red
    """

    def __init__(self, spi, n, *, brightness=1.0, auto_write=True, pixel_order=BGR):
        self._spi = spi
        self._n = n
        # Supply one extra clock cycle for each two pixels in the strip.
        self.end_header_size = n // 16
        if n % 16 != 0:
            self.end_header_size += 1
        self._buf = bytearray(n * 4 + START_HEADER_SIZE + self.end_header_size)
        self.end_header_index = len(self._buf) - self.end_header_size
        self.pixel_order = pixel_order
        # Four empty bytes to start.
        for i in range(START_HEADER_SIZE):
            self._buf[i] = 0x00
        # Mark the beginnings of each pixel.
        for i in range(START_HEADER_SIZE, self.end_header_index, 4):
            self._buf[i] = 0xFF
        # 0xff bytes at the end.
        for i in range(self.end_header_index, len(self._buf)):
            self._buf[i] = 0xFF
        self._brightness = 1.0
        # Set auto_write to False temporarily so brightness setter does _not_
        # call show() while in __init__.
        self.auto_write = False
        self.brightness = brightness
        self.auto_write = auto_write

    def deinit(self):
        """Blank out the DotStars and release the resources."""
        self.auto_write = False
        for i in range(START_HEADER_SIZE, self.end_header_index):
            if i % 4 != 0:
                self._buf[i] = 0
        self.show()
        if self._spi:
            self._spi.deinit()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.deinit()

    def __repr__(self):
        return "[" + ", ".join([str(x) for x in self]) + "]"

    def _set_item(self, index, value):
        """
        value can be one of three things:
                a (r,g,b) list/tuple
                a (r,g,b, brightness) list/tuple
                a single, longer int that contains RGB values, like 0xFFFFFF
            brightness, if specified should be a float 0-1

        Set a pixel value. You can set per-pixel brightness here, if it's not passed it
        will use the max value for pixel brightness value, which is a good default.

        Important notes about the per-pixel brightness - it's accomplished by
        PWMing the entire output of the LED, and that PWM is at a much
        slower clock than the rest of the LEDs. This can cause problems in
        Persistence of Vision Applications
        """

        offset = index * 4 + START_HEADER_SIZE
        rgb = value
        if isinstance(value, int):
            rgb = (value >> 16, (value >> 8) & 0xFF, value & 0xFF)

        if len(rgb) == 4:
            brightness = value[3]
            # Ignore value[3] below.
        else:
            brightness = 1

        # LED startframe is three "1" bits, followed by 5 brightness bits
        # then 8 bits for each of R, G, and B. The order of those 3 are configurable and
        # vary based on hardware
        # same as math.ceil(brightness * 31) & 0b00011111
        # Idea from https://www.codeproject.com/Tips/700780/Fast-floor-ceiling-functions
        brightness_byte = 32 - int(32 - brightness * 31) & 0b00011111
        self._buf[offset] = brightness_byte | LED_START
        self._buf[offset + 1] = rgb[self.pixel_order[0]]
        self._buf[offset + 2] = rgb[self.pixel_order[1]]
        self._buf[offset + 3] = rgb[self.pixel_order[2]]

    def __setitem__(self, index, val):
        if isinstance(index, slice):
            start, stop, step = index.indices(self._n)
            length = stop - start
            if step != 0:
                # same as math.ceil(length / step)
                # Idea from https://fizzbuzzer.com/implement-a-ceil-function/
                length = (length + step - 1) // step
            if len(val) != length:
                raise ValueError("Slice and input sequence size do not match.")
            for val_i, in_i in enumerate(range(start, stop, step)):
                self._set_item(in_i, val[val_i])
        else:
            self._set_item(index, val)

        if self.auto_write:
            self.show()

    def __getitem__(self, index):
        if isinstance(index, slice):
            out = []
            for in_i in range(*index.indices(self._n)):
                out.append(
                    tuple(self._buf[in_i * 4 + (3 - i) + START_HEADER_SIZE] for i in range(3))
                )
            return out
        if index < 0:
            index += len(self)
        if index >= self._n or index < 0:
            raise IndexError
        offset = index * 4
        return tuple(self._buf[offset + (3 - i) + START_HEADER_SIZE] for i in range(3))

    def __len__(self):
        return self._n

    @property
    def brightness(self):
        """Overall brightness of the pixel"""
        return self._brightness

    @brightness.setter
    def brightness(self, brightness):
        self._brightness = min(max(brightness, 0.0), 1.0)
        if self.auto_write:
            self.show()

    def fill(self, color):
        """Colors all pixels the given ***color***."""
        auto_write = self.auto_write
        self.auto_write = False
        for i in range(self._n):
            self[i] = color
        if auto_write:
            self.show()
        self.auto_write = auto_write

    def show(self):
        """Shows the new colors on the pixels themselves if they haven't already
        been autowritten.

        The colors may or may not be showing after this function returns because
        it may be done asynchronously."""
        # Create a second output buffer if we need to compute brightness
        buf = self._buf
        if self.brightness < 1.0:
            buf = bytearray(self._buf)
            # Four empty bytes to start.
            for i in range(START_HEADER_SIZE):
                buf[i] = 0x00
            for i in range(START_HEADER_SIZE, self.end_header_index):
                buf[i] = self._buf[i] if i % 4 == 0 else int(self._buf[i] * self._brightness)
            # Four 0xff bytes at the end.
            for i in range(self.end_header_index, len(buf)):
                buf[i] = 0xFF

        if self._spi:
            self._spi.write(buf)



# Test accuracy of sleep within a thread.

import time
import _thread


def sleep(t, valid_range):
    t0 = time.time_ns()
    time.sleep(t)
    dt_ms = (time.time_ns() - t0) // 1_000_000
    if dt_ms in valid_range:
        print("dt in range", t)
    else:
        print("dt not in range:", dt_ms)


def thread_entry():
    lock.acquire()
    print("thread start")
    sleep(0.2, range(180, 400))
    print("thread end")


lock = _thread.allocate_lock()
lock.acquire()
_thread.start_new_thread(thread_entry, ())

print("main start")
lock.release()
sleep(0.5, range(480, 800))
print("main end")



# basic multiplication
print('0' * 5)

# check negative, 0, positive; lhs and rhs multiplication
for i in (-4, -2, 0, 2, 4):
    print(i * '12')
    print('12' * i)

# check that we don't modify existing object
a = '123'
c = a * 3
print(a, c)

# check inplace multiplication
a = '456'
a *= 3
print(a)

# check reverse inplace multiplication
a = 3
a *= '789'
print(a)



try:
    from collections import deque
except ImportError:
    print("SKIP")
    raise SystemExit

d = deque((), 10)

d.append(1)
d.append(2)
d.append(3)

# Index slicing for reads is not supported in CPython
try:
    d[0:1]
except TypeError:
    print("TypeError")

# Index slicing for writes is not supported in CPython
try:
    d[0:1] = (-1, -2)
except TypeError:
    print("TypeError")

# Index slicing for writes is not supported in CPython
try:
    del d[0:1]
except TypeError:
    print("TypeError")



#!/usr/bin/env python
#
# Takes a directory of files and zips them up (as uncompressed files).
# This then gets converted into a C data structure which can be read
# like a filesystem at runtime.
#
# This is somewhat like frozen modules in python, but allows arbitrary files
# to be used.

from __future__ import print_function

import argparse
import os
import subprocess
import sys
import types


def create_zip(zip_filename, zip_dir):
    abs_zip_filename = os.path.abspath(zip_filename)
    save_cwd = os.getcwd()
    os.chdir(zip_dir)
    if os.path.exists(abs_zip_filename):
        os.remove(abs_zip_filename)
    subprocess.check_call(["zip", "-0", "-r", "-D", abs_zip_filename, "."])
    os.chdir(save_cwd)


def create_c_from_file(c_filename, zip_filename):
    with open(zip_filename, "rb") as zip_file:
        with open(c_filename, "wb") as c_file:
            print("#include <stdint.h>", file=c_file)
            print("", file=c_file)
            print("const uint8_t memzip_data[] = {", file=c_file)
            while True:
                buf = zip_file.read(16)
                if not buf:
                    break
                print("   ", end="", file=c_file)
                for byte in buf:
                    if isinstance(byte, types.StringType):
                        print(" 0x{:02x},".format(ord(byte)), end="", file=c_file)
                    else:
                        print(" 0x{:02x},".format(byte), end="", file=c_file)
                print("", file=c_file)
            print("};", file=c_file)


def main():
    parser = argparse.ArgumentParser(
        prog="make-memzip.py",
        usage="%(prog)s [options] [command]",
        description="Generates a C source memzip file.",
    )
    parser.add_argument(
        "-z",
        "--zip-file",
        dest="zip_filename",
        help="Specifies the name of the created zip file.",
        default="memzip_files.zip",
    )
    parser.add_argument(
        "-c",
        "--c-file",
        dest="c_filename",
        help="Specifies the name of the created C source file.",
        default="memzip_files.c",
    )
    parser.add_argument(dest="source_dir", default="memzip_files")
    args = parser.parse_args(sys.argv[1:])

    print("args.zip_filename =", args.zip_filename)
    print("args.c_filename =", args.c_filename)
    print("args.source_dir =", args.source_dir)

    create_zip(args.zip_filename, args.source_dir)
    create_c_from_file(args.c_filename, args.zip_filename)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# This file is part of the MicroPython project, http://micropython.org/
#
# The MIT License (MIT)
#
# Copyright (c) 2023 Jim Mussared
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# This script makes a CPython-compatible package from a micropython-lib package
# with a pyproject.toml that can be built (via hatch) and deployed to PyPI.
# Requires that the project sets the pypi_publish= kwarg in its metadata().

# Usage:
# ./tools/makepyproject.py --output /tmp/foo micropython/foo
# python -m build /tmp/foo
# python -m twine upload /tmp/foo/dist/*.whl

from email.utils import parseaddr
import os
import re
import shutil
import sys

from build import error_color, ensure_path_exists


DEFAULT_AUTHOR = "micropython-lib <contact@micropython.org>"
DEFAULT_LICENSE = "MIT"


def quoted_escape(s):
    return s.replace('"', '\\"')


def build(manifest_path, output_path):
    import manifestfile

    if not manifest_path.endswith(".py"):
        # Allow specifying either the directory or the manifest file explicitly.
        manifest_path = os.path.join(manifest_path, "manifest.py")

    print("Generating pyproject for {} in {}...".format(manifest_path, output_path))

    toml_path = os.path.join(output_path, "pyproject.toml")
    ensure_path_exists(toml_path)

    path_vars = {
        "MPY_LIB_DIR": os.path.abspath(os.path.join(os.path.dirname(__file__), "..")),
    }

    # .../foo/manifest.py -> foo
    package_name = os.path.basename(os.path.dirname(manifest_path))

    # Compile the manifest.
    manifest = manifestfile.ManifestFile(manifestfile.MODE_PYPROJECT, path_vars)
    manifest.execute(manifest_path)

    # If a package doesn't have a pypi name, then assume it isn't intended to
    # be publishable.
    if not manifest.metadata().pypi_publish:
        print(error_color("Error:"), package_name, "doesn't have a pypi_publish name.")
        sys.exit(1)

    # These should be in all packages eventually.
    if not manifest.metadata().version:
        print(error_color("Error:"), package_name, "doesn't have a version.")
        sys.exit(1)
    if not manifest.metadata().description:
        print(error_color("Error:"), package_name, "doesn't have a description.")
        sys.exit(1)

    # This is the root path of all .py files that are copied. We ensure that
    # they all match.
    top_level_package = None

    for result in manifest.files():
        # This isn't allowed in micropython-lib anyway.
        if result.file_type != manifestfile.FILE_TYPE_LOCAL:
            print(error_color("Error:"), "Non-local file not supported.", file=sys.stderr)
            sys.exit(1)

        # "foo/bar/baz.py" --> "foo"
        # "baz.py" --> ""
        result_package = os.path.split(result.target_path)[0]

        if not result_package:
            # This is a standalone .py file.
            print(
                error_color("Error:"),
                "Unsupported single-file module: {}".format(result.target_path),
                file=sys.stderr,
            )
            sys.exit(1)
        if top_level_package and result_package != top_level_package:
            # This likely suggests that something needs to use require(..., pypi="...").
            print(
                error_color("Error:"),
                "More than one top-level package: {}, {}.".format(
                    result_package, top_level_package
                ),
                file=sys.stderr,
            )
            sys.exit(1)
        top_level_package = result_package

        # Tag each file with the package metadata and copy the .py directly.
        with manifestfile.tagged_py_file(result.full_path, result.metadata) as tagged_path:
            dest_path = os.path.join(output_path, result.target_path)
            ensure_path_exists(dest_path)
            shutil.copyfile(tagged_path, dest_path)

    # Copy README.md if it exists
    readme_path = os.path.join(os.path.dirname(manifest_path), "README.md")
    readme_toml = ""
    if os.path.exists(readme_path):
        shutil.copyfile(readme_path, os.path.join(output_path, "README.md"))
        readme_toml = 'readme = "README.md"'

    # Apply default author and license, otherwise use the package metadata.
    license_toml = 'license = {{ text = "{}" }}'.format(
        quoted_escape(manifest.metadata().license or DEFAULT_LICENSE)
    )
    author_name, author_email = parseaddr(manifest.metadata().author or DEFAULT_AUTHOR)
    author_toml = 'authors = [ {{ name = "{}", email = "{}"}} ]'.format(
        quoted_escape(author_name), quoted_escape(author_email)
    )

    # Write pyproject.toml.
    with open(toml_path, "w") as toml_file:
        print("# Generated by makepyproject.py", file=toml_file)

        print(
            """
[build-system]
requires = [
    "hatchling"
]
build-backend = "hatchling.build"
""",
            file=toml_file,
        )

        print(
            """
[project]
name = "{}"
description = "{}"
{}
{}
version = "{}"
dependencies = [{}]
urls = {{ Homepage = "https://github.com/micropython/micropython-lib" }}
{}
""".format(
                quoted_escape(manifest.metadata().pypi_publish),
                quoted_escape(manifest.metadata().description),
                author_toml,
                license_toml,
                quoted_escape(manifest.metadata().version),
                ", ".join('"{}"'.format(quoted_escape(r)) for r in manifest.pypi_dependencies()),
                readme_toml,
            ),
            file=toml_file,
        )

        print(
            """
[tool.hatch.build]
packages = ["{}"]
""".format(top_level_package),
            file=toml_file,
        )

    print("Done.")


def main():
    import argparse

    cmd_parser = argparse.ArgumentParser(
        description="Generate a project that can be pushed to PyPI."
    )
    cmd_parser.add_argument("--output", required=True, help="output directory")
    cmd_parser.add_argument("--micropython", default=None, help="path to micropython repo")
    cmd_parser.add_argument("manifest", help="input package path")
    args = cmd_parser.parse_args()

    if args.micropython:
        sys.path.append(os.path.join(args.micropython, "tools"))  # for manifestfile

    build(args.manifest, args.output)


if __name__ == "__main__":
    main()



# test that emergency exceptions work

import micropython
import sys

try:
    import io
except ImportError:
    print("SKIP")
    raise SystemExit

# some ports need to allocate heap for the emg exc
try:
    micropython.alloc_emergency_exception_buf(256)
except AttributeError:
    pass


def f():
    micropython.heap_lock()
    try:
        raise ValueError(1)
    except ValueError as er:
        exc = er
    micropython.heap_unlock()

    # print the exception
    buf = io.StringIO()
    sys.print_exception(exc, buf)
    for l in buf.getvalue().split("\n"):
        if l.startswith("  File "):
            print(l.split('"')[2])
        else:
            print(l)


f()



"""
The MIT License (MIT)

Copyright (c) 2013, 2014 Damien P. George
Copyright (c) 2022-2023 Ibrahim Abdelkader <iabdalkader@openmv.io>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.


LSM9DS1 - 9DOF inertial sensor of STMicro driver for MicroPython.
The sensor contains an accelerometer / gyroscope / magnetometer
Uses the internal FIFO to store up to 16 gyro/accel data, use the iter_accel_gyro generator to access it.

Example usage:

import time
from lsm9ds1 import LSM9DS1
from machine import Pin, I2C

imu = LSM9DS1(I2C(1, scl=Pin(15), sda=Pin(14)))

while (True):
    #for g,a in imu.iter_accel_gyro(): print(g,a)    # using fifo
    print('Accelerometer: x:{:>8.3f} y:{:>8.3f} z:{:>8.3f}'.format(*imu.accel()))
    print('Magnetometer:  x:{:>8.3f} y:{:>8.3f} z:{:>8.3f}'.format(*imu.magnet()))
    print('Gyroscope:     x:{:>8.3f} y:{:>8.3f} z:{:>8.3f}'.format(*imu.gyro()))
    print("")
    time.sleep_ms(100)
"""
import array
from micropython import const


_WHO_AM_I = const(0xF)
_CTRL_REG1_G = const(0x10)
_INT_GEN_SRC_G = const(0x14)
_OUT_TEMP = const(0x15)
_OUT_G = const(0x18)
_CTRL_REG4_G = const(0x1E)
_STATUS_REG = const(0x27)
_OUT_XL = const(0x28)
_FIFO_CTRL_REG = const(0x2E)
_FIFO_SRC = const(0x2F)
_OFFSET_REG_X_M = const(0x05)
_CTRL_REG1_M = const(0x20)
_OUT_M = const(0x28)
_ACCEL_SCALE = const((2, 16, 4, 8))
_GYRO_SCALE = const((245, 500, 2000))
_MAGNET_SCALE = const((4, 8, 12, 16))
_ODR_IMU = const((0, 14.9, 59.5, 119, 238, 476, 952))
_ODR_MAGNET = const((0.625, 1.25, 2.5, 5, 10, 20, 40, 80))


class LSM9DS1:
    def __init__(
        self,
        bus,
        address_imu=0x6B,
        address_magnet=0x1E,
        gyro_odr=952,
        gyro_scale=245,
        accel_odr=952,
        accel_scale=4,
        magnet_odr=80,
        magnet_scale=4,
    ):
        """Initalizes Gyro, Accelerometer and Magnetometer.
        bus: IMU bus
        address_imu: IMU I2C address.
        address_magnet: Magnetometer I2C address.
        gyro_odr: (0, 14.9Hz, 59.5Hz, 119Hz, 238Hz, 476Hz, 952Hz)
        gyro_scale: (245dps, 500dps, 2000dps )
        accel_odr: (0, 14.9Hz, 59.5Hz, 119Hz, 238Hz, 476Hz, 952Hz)
        accel_scale: (+/-2g, +/-4g, +/-8g, +-16g)
        magnet_odr: (0.625Hz, 1.25Hz, 2.5Hz, 5Hz, 10Hz, 20Hz, 40Hz, 80Hz)
        magnet_scale: (+/-4, +/-8, +/-12, +/-16)
        """
        self.bus = bus
        self.address_imu = address_imu
        self.address_magnet = address_magnet

        # Sanity checks
        if gyro_odr not in _ODR_IMU:
            raise ValueError("Invalid gyro sampling rate: %d" % gyro_odr)
        if gyro_scale not in _GYRO_SCALE:
            raise ValueError("Invalid gyro scaling: %d" % gyro_scale)

        if accel_odr not in _ODR_IMU:
            raise ValueError("Invalid accelerometer sampling rate: %d" % accel_odr)
        if accel_scale not in _ACCEL_SCALE:
            raise ValueError("Invalid accelerometer scaling: %d" % accel_scale)

        if magnet_odr not in _ODR_MAGNET:
            raise ValueError("Invalid magnet sampling rate: %d" % magnet_odr)
        if magnet_scale not in _MAGNET_SCALE:
            raise ValueError("Invalid magnet scaling: %d" % magnet_scale)

        if (self.magent_id() != b"=") or (self.gyro_id() != b"h"):
            raise OSError(
                "Invalid LSM9DS1 device, using address {}/{}".format(address_imu, address_magnet)
            )

        mv = memoryview(bytearray(6))

        # Configure Gyroscope.
        mv[0] = (_ODR_IMU.index(gyro_odr) << 5) | ((_GYRO_SCALE.index(gyro_scale)) << 3)
        mv[1:4] = b"\x00\x00\x00"
        self.bus.writeto_mem(self.address_imu, _CTRL_REG1_G, mv[:5])

        # Configure Accelerometer
        mv[0] = 0x38  # ctrl4 - enable x,y,z, outputs, no irq latching, no 4D
        mv[1] = 0x38  # ctrl5 - enable all axes, no decimation
        # ctrl6 - set scaling and sample rate of accel
        mv[2] = (_ODR_IMU.index(accel_odr) << 5) | ((_ACCEL_SCALE.index(accel_scale)) << 3)
        mv[3] = 0x00  # ctrl7 - leave at default values
        mv[4] = 0x4  # ctrl8 - leave at default values
        mv[5] = 0x2  # ctrl9 - FIFO enabled
        self.bus.writeto_mem(self.address_imu, _CTRL_REG4_G, mv)

        # fifo: use continous mode (overwrite old data if overflow)
        self.bus.writeto_mem(self.address_imu, _FIFO_CTRL_REG, b"\x00")
        self.bus.writeto_mem(self.address_imu, _FIFO_CTRL_REG, b"\xc0")

        # Configure Magnetometer
        mv[0] = 0x40 | (magnet_odr << 2)  # ctrl1: high performance mode
        mv[1] = _MAGNET_SCALE.index(magnet_scale) << 5  # ctrl2: scale, normal mode, no reset
        mv[2] = 0x00  # ctrl3: continous conversion, no low power, I2C
        mv[3] = 0x08  # ctrl4: high performance z-axis
        mv[4] = 0x00  # ctr5: no fast read, no block update
        self.bus.writeto_mem(self.address_magnet, _CTRL_REG1_M, mv[:5])

        self.gyro_scale = 32768 / gyro_scale
        self.accel_scale = 32768 / accel_scale
        self.scale_factor_magnet = 32768 / ((_MAGNET_SCALE.index(magnet_scale) + 1) * 4)

        # Allocate scratch buffer for efficient conversions and memread op's
        self.scratch_int = array.array("h", [0, 0, 0])

    def calibrate_magnet(self, offset):
        """
        offset is a magnet vector that will be subtracted by the magnetometer
        for each measurement. It is written to the magnetometer's offset register
        """
        import struct

        offset = [int(i * self.scale_factor_magnet) for i in offset]
        self.bus.writeto_mem(self.address_magnet, _OFFSET_REG_X_M, struct.pack("<HHH", offset))

    def gyro_id(self):
        return self.bus.readfrom_mem(self.address_imu, _WHO_AM_I, 1)

    def magent_id(self):
        return self.bus.readfrom_mem(self.address_magnet, _WHO_AM_I, 1)

    def magnet(self):
        """Returns magnetometer vector in gauss.
        raw_values: if True, the non-scaled adc values are returned
        """
        mv = memoryview(self.scratch_int)
        f = self.scale_factor_magnet
        self.bus.readfrom_mem_into(self.address_magnet, _OUT_M | 0x80, mv)
        return (mv[0] / f, mv[1] / f, mv[2] / f)

    def gyro(self):
        """Returns gyroscope vector in degrees/sec."""
        mv = memoryview(self.scratch_int)
        f = self.gyro_scale
        self.bus.readfrom_mem_into(self.address_imu, _OUT_G | 0x80, mv)
        return (mv[0] / f, mv[1] / f, mv[2] / f)

    def accel(self):
        """Returns acceleration vector in gravity units (9.81m/s^2)."""
        mv = memoryview(self.scratch_int)
        f = self.accel_scale
        self.bus.readfrom_mem_into(self.address_imu, _OUT_XL | 0x80, mv)
        return (mv[0] / f, mv[1] / f, mv[2] / f)

    def iter_accel_gyro(self):
        """A generator that returns tuples of (gyro,accelerometer) data from the fifo."""
        while True:
            fifo_state = int.from_bytes(
                self.bus.readfrom_mem(self.address_imu, _FIFO_SRC, 1), "big"
            )
            if fifo_state & 0x3F:
                # print("Available samples=%d" % (fifo_state & 0x1f))
                yield self.gyro(), self.accel()
            else:
                break



class CtxMgr:

    def __enter__(self):
        print("__enter__")
        return self

    def __exit__(self, a, b, c):
        print("__exit__", repr(a), repr(b))

for i in range(5):
    print(i)
    with CtxMgr():
        if i == 3:
            break



# MicroPython LoRa SX127x driver
# MIT license; Copyright (c) 2023 Angus Gratton
#
# LoRa is a registered trademark or service mark of Semtech Corporation or its affiliates.
#
# In comments, abbreviation "DS" = Semtech SX1276/77/78/79 Datasheet rev 7 (May 2020)
from micropython import const
from .modem import BaseModem, ConfigError, RxPacket, _clamp, _flag
from machine import Pin
import struct
import time

# Set _DEBUG to const(True) to print all register reads and writes, and current register values
# even when an update isn't needed. Plus a few additional pieces of information.
_DEBUG = const(False)

_WRITE_REG_BIT = const(1 << 7)

# Registers and fields as bytecode-zerocost constants
#
# Where possible names are direct from DS section 4.4
# (This means some names are slightly inconsistent, as per datasheet...)

_REG_FIFO = const(0x00)

_REG_OPMODE = const(0x01)

_OPMODE_LONGRANGEMODE_LORA = const(1 << 7)
_OPMODE_LONGRANGEMODE_FSK_OOK = const(0)
_OPMODE_MODE_MASK = const(0x7)
_OPMODE_MODE_SLEEP = const(0x0)
_OPMODE_MODE_STDBY = const(0x1)
_OPMODE_MODE_FSTX = const(0x2)  # Frequency synthesis (TX)
_OPMODE_MODE_TX = const(0x3)
_OPMODE_MODE_FSRX = const(0x4)  # Frequency synthesis (RX)
_OPMODE_MODE_RX_CONTINUOUS = const(0x5)
_OPMODE_MODE_RX_SINGLE = const(0x6)
_OPMODE_MODE_CAD = const(0x7)  # Channel Activity Detection

_REG_FR_MSB = const(0x06)
_REG_FR_MID = const(0x07)
_REG_FR_LSB = const(0x08)

_REG_PA_CONFIG = const(0x09)

_PA_CONFIG_PASELECT_PA_BOOST_PIN = const(1 << 7)
_PA_CONFIG_PASELECT_RFO_PIN = const(0x0)
_PA_CONFIG_MAXPOWER_SHIFT = const(0x4)
_PA_CONFIG_MAXPOWER_MASK = const(0x7)
_PA_CONFIG_OUTPUTPOWER_SHIFT = const(0)
_PA_CONFIG_OUTPUTPOWER_MASK = const(0xF)

_REG_PA_RAMP = const(0x0A)
_PA_RAMP_MASK = const(0x0F)

_REG_LNA = const(0x0C)

_LNA_GAIN_MASK = const(0x7)
_LNA_GAIN_SHIFT = const(5)

_LNA_BOOST_HF_MASK = 0x3
_LNA_BOOST_HF_SHIFT = 0x0

_REG_FIFO_ADDR_PTR = const(0x0D)
_REG_FIFO_TX_BASE_ADDR = const(0x0E)
_REG_FIFO_RX_BASE_ADDR = const(0x0F)
_REG_FIFO_RX_CURRENT_ADDR = const(0x10)

_REG_IRQ_FLAGS_MASK = const(0x11)
_REG_IRQ_FLAGS = const(0x12)

# IRQ mask bits are the same as the IRQ flag bits
_IRQ_RX_TIMEOUT = const(1 << 7)
_IRQ_RX_DONE = const(1 << 6)
_IRQ_PAYLOAD_CRC_ERROR = const(1 << 5)
_IRQ_VALID_HEADER = const(1 << 4)
_IRQ_TX_DONE = const(1 << 3)
_IRQ_CAD_DONE = const(1 << 2)
_IRQ_FHSS_CHANGE_CHANNEL = const(1 << 1)
_IRQ_CAD_DETECTED = const(1 << 0)

_REG_RX_NB_BYTES = const(0x13)
_REG_RX_HEADER_CNT_VALUE_MSB = const(0x14)
_REG_RX_HEADER_CNT_VALUE_LSB = const(0x13)
_REG_RX_PACKET_CNT_VALUE_MSB = const(0x16)
_REG_RX_PACKET_CNT_VALUE_LSB = const(0x17)

_REG_MODEM_STAT = const(0x18)
_MODEM_STAT_RX_CODING_RATE_MASK = const(0xE)
_MODEM_STAT_RX_CODING_RATE_SHIFT = const(5)
_MODEM_STAT_MODEM_CLEAR = const(1 << 4)
_MODEM_STAT_HEADER_INFO_VALID = const(1 << 3)
_MODEM_STAT_RX_ONGOING = const(1 << 2)
_MODEM_STAT_SIGNAL_SYNC = const(1 << 1)  # Signal synchronized
_MODEM_STAT_SIGNAL_DET = const(1 << 0)  # Signal detected

_REG_PKT_SNR_VAL = const(0x19)
_REG_PKT_RSSI_VAL = const(0x1A)
_REG_RSSI_VAL = const(0x1B)

_REG_HOP_CHANNEL = const(0x1C)
_HOP_CHANNEL_PLL_TIMEOUT = const(1 << 7)
_HOP_CHANNEL_CRC_ON_PAYLOAD = const(1 << 6)
_HOP_CHANNEL_FHSS_PRESENT_CHANNEL_MASK = const(0x1F)

_REG_MODEM_CONFIG1 = const(0x1D)
_MODEM_CONFIG1_BW_MASK = const(0xF)
_MODEM_CONFIG1_BW_SHIFT = const(4)
_MODEM_CONFIG1_BW7_8 = const(0x0)
_MODEM_CONFIG1_BW10_4 = const(0x1)
_MODEM_CONFIG1_BW15_6 = const(0x2)
_MODEM_CONFIG1_BW20_8 = const(0x3)
_MODEM_CONFIG1_BW31_25 = const(0x4)
_MODEM_CONFIG1_BW41_7 = const(0x5)
_MODEM_CONFIG1_BW62_5 = const(0x6)
_MODEM_CONFIG1_BW125 = const(0x7)
_MODEM_CONFIG1_BW250 = const(0x8)  # not supported in lower band (169MHz)
_MODEM_CONFIG1_BW500 = const(0x9)  # not supported in lower band (169MHz)
_MODEM_CONFIG1_CODING_RATE_MASK = const(0x7)
_MODEM_CONFIG1_CODING_RATE_SHIFT = const(1)
_MODEM_CONFIG1_CODING_RATE_45 = const(0b001)
_MODEM_CONFIG1_CODING_RATE_46 = const(0b010)
_MODEM_CONFIG1_CODING_RATE_47 = const(0b011)
_MODEM_CONFIG1_CODING_RATE_48 = const(0b100)
_MODEM_CONFIG1_IMPLICIT_HEADER_MODE_ON = const(1 << 0)

_REG_MODEM_CONFIG2 = const(0x1E)
_MODEM_CONFIG2_SF_MASK = const(0xF)  # Spreading Factor
_MODEM_CONFIG2_SF_SHIFT = const(4)
# SF values are integers 6-12 for SF6-SF12, so skipping constants for these
_MODEM_CONFIG2_SF_MIN = const(6)  # inclusive
_MODEM_CONFIG2_SF_MAX = const(12)  # inclusive

_MODEM_CONFIG2_TX_CONTINUOUS = const(1 << 3)
_MODEM_CONFIG2_RX_PAYLOAD_CRC_ON = const(1 << 2)
_MODEM_CONFIG2_SYMB_TIMEOUT_MSB_MASK = 0x3

_REG_SYMB_TIMEOUT_LSB = const(0x1F)

_REG_PREAMBLE_LEN_MSB = const(0x20)
_REG_PREAMBLE_LEN_LSB = const(0x21)

_REG_PAYLOAD_LEN = const(0x22)  # Only for implicit header mode & TX
_REG_MAX_PAYLOAD_LEN = const(0x23)

_REG_HOP_PERIOD = const(0x24)

_REG_FIFO_TXBYTE_ADDR = const(0x25)

_REG_MODEM_CONFIG3 = const(0x26)
_MODEM_CONFIG3_AGC_ON = const(1 << 2)
_MODEM_CONFIG3_LOW_DATA_RATE_OPTIMIZE = const(1 << 3)

_REG_DETECT_OPTIMIZE = const(0x31)
_DETECT_OPTIMIZE_AUTOMATIC_IF_ON = const(
    1 << 7
)  # Bit should be cleared after reset, as per errata
_DETECT_OPTIMIZE_MASK = 0x7
_DETECT_OPTIMIZE_SF6 = const(0x05)
_DETECT_OPTIMIZE_OTHER = const(0x03)

# RegInvertIQ is not correctly documented in DS Rev 7 (May 2020).
#
# The correct behaviour for interoperability with other LoRa devices is as
# written here:
# https://github.com/eclipse/upm/blob/master/src/sx1276/sx1276.cxx#L1310
#
# Same as used in the Semtech mbed driver, here:
# https://github.com/ARMmbed/mbed-semtech-lora-rf-drivers/blob/master/SX1276/SX1276_LoRaRadio.cpp#L778
# https://github.com/ARMmbed/mbed-semtech-lora-rf-drivers/blob/master/SX1276/registers/sx1276Regs-LoRa.h#L443
#
# Specifically:
# - The TX bit in _REG_INVERT_IQ is opposite to what's documented in the datasheet
#   (0x01 normal, 0x00 inverted)
# - The RX bit in _REG_INVERT_IQ is as documented in the datasheet (0x00 normal, 0x40 inverted)
# - When enabling LoRa mode, the default register value becomes 0x27 (normal RX & TX)
#   rather than the documented power-on value of 0x26.
_REG_INVERT_IQ = const(0x33)
_INVERT_IQ_RX = const(1 << 6)
_INVERT_IQ_TX_OFF = const(1 << 0)

_REG_DETECTION_THRESHOLD = const(0x37)
_DETECTION_THRESHOLD_SF6 = const(0x0C)
_DETECTION_THRESHOLD_OTHER = const(0x0A)  # SF7 to SF12

_REG_SYNC_WORD = const(0x39)

_REG_FSKOOK_IMAGE_CAL = const(0x3B)  # NOTE: Only accessible in FSK/OOK mode
_IMAGE_CAL_START = const(1 << 6)
_IMAGE_CAL_RUNNING = const(1 << 5)
_IMAGE_CAL_AUTO = const(1 << 7)

_REG_INVERT_IQ2 = const(0x3B)
_INVERT_IQ2_ON = const(0x19)
_INVERT_IQ2_OFF = const(0x1D)

_REG_DIO_MAPPING1 = const(0x40)
_DIO0_MAPPING_MASK = const(0x3)
_DIO0_MAPPING_SHIFT = const(6)
_DIO1_MAPPING_MASK = const(0x3)
_DIO1_MAPPING_SHIFT = const(4)
_DIO2_MAPPING_MASK = const(0x3)
_DIO2_MAPPING_SHIFT = const(2)
_DIO3_MAPPING_MASK = const(0x3)
_DIO3_MAPPING_SHIFT = const(0)

_REG_DIO_MAPPING2 = const(0x41)
_DIO4_MAPPING_MASK = const(0x3)
_DIO4_MAPPING_SHIFT = const(6)
_DIO5_MAPPING_MASK = const(0x3)
_DIO5_MAPPING_SHIFT = const(4)

_REG_PA_DAC = const(0x4D)
_PA_DAC_DEFAULT_VALUE = const(0x84)  # DS 3.4.3 High Power +20 dBm Operation
_PA_DAC_HIGH_POWER_20DBM = const(0x87)

_REG_VERSION = const(0x42)

# IRQs the driver masks in when receiving
_IRQ_DRIVER_RX_MASK = const(
    _IRQ_RX_DONE | _IRQ_RX_TIMEOUT | _IRQ_VALID_HEADER | _IRQ_PAYLOAD_CRC_ERROR
)


class _SX127x(BaseModem):
    # Don't instantiate this class directly, instantiate either lora.SX1276,
    #  lora.SX1277, lora.SX1278, lora.SX1279, or lora.AsyncSX1276,
    #  lora.AsyncSX1277, lora.AsyncSX1278, lora.AsyncSX1279 as applicable.

    # common IRQ masks used by the base class functions
    _IRQ_RX_COMPLETE = _IRQ_RX_DONE | _IRQ_RX_TIMEOUT
    _IRQ_TX_COMPLETE = _IRQ_TX_DONE

    def __init__(self, spi, cs, dio0=None, dio1=None, reset=None, lora_cfg=None, ant_sw=None):
        super().__init__(ant_sw)

        self._buf1 = bytearray(1)  # shared small buffers
        self._buf2 = bytearray(2)
        self._spi = spi
        self._cs = cs

        self._dio0 = dio0
        self._dio1 = dio1

        cs.init(Pin.OUT, value=1)

        if dio0:
            dio0.init(Pin.IN)
            dio0.irq(self._radio_isr, trigger=Pin.IRQ_RISING)
            if dio1:
                dio1.init(Pin.IN)
                dio1.irq(self._radio_isr, trigger=Pin.IRQ_RISING)

        # Configuration settings that need to be tracked by the driver
        # Note: a number of these are set in the base class constructor
        self._pa_boost = False

        if reset:
            # If the user supplies a reset pin argument, reset the radio
            reset.init(Pin.OUT, value=0)
            time.sleep_ms(1)
            reset(1)
            time.sleep_ms(5)

        version = self._reg_read(_REG_VERSION)
        if version != 0x12:
            raise RuntimeError("Unexpected silicon version {}".format(version))

        # wake the radio and enable LoRa mode if it's not already set
        self._set_mode(_OPMODE_MODE_STDBY)

        if lora_cfg:
            self.configure(lora_cfg)

    def configure(self, lora_cfg):
        if self._rx is not False:
            raise RuntimeError("Receiving")

        # Set frequency
        if "freq_khz" in lora_cfg:
            # Assuming F(XOSC)=32MHz (datasheet both implies this value can be different, and
            # specifies it shouldn't be different!)
            self._rf_freq_hz = int(lora_cfg["freq_khz"] * 1000)
            fr_val = self._rf_freq_hz * 16384 // 1000_000
            buf = bytes([fr_val >> 16, (fr_val >> 8) & 0xFF, fr_val & 0xFF])
            self._reg_write(_REG_FR_MSB, buf)

        # Turn on/off automatic image re-calibration if temperature changes. May lead to dropped
        # packets if enabled.
        if "auto_image_cal" in lora_cfg:
            self._set_mode(_OPMODE_MODE_STDBY, False)  # Disable LoRa mode to access FSK/OOK
            self._reg_update(
                _REG_FSKOOK_IMAGE_CAL,
                _IMAGE_CAL_AUTO,
                _flag(_IMAGE_CAL_AUTO, lora_cfg["auto_image_cal"]),
            )
            self._set_mode(_OPMODE_MODE_STDBY)  # Switch back to LoRa mode

        # Note: Common pattern below is to generate a new register value and an update_mask,
        # and then call self._reg_update(). self._reg_update() is a
        # no-op if update_mask==0 (no bits to change).

        # Update _REG_PA_CONFIG
        pa_config = 0x0
        update_mask = 0x0

        # Ref DS 3.4.2 "RF Power Amplifiers"
        if "tx_ant" in lora_cfg:
            self._pa_boost = lora_cfg["tx_ant"].upper() == "PA_BOOST"
            pa_boost_bit = (
                _PA_CONFIG_PASELECT_PA_BOOST_PIN if self._pa_boost else _PA_CONFIG_PASELECT_RFO_PIN
            )
            pa_config |= pa_boost_bit
            update_mask |= pa_boost_bit
            if not self._pa_boost:
                # When using RFO, _REG_PA_DAC can keep default value always
                # (otherwise, it's set when output_power is set in next block)
                self._reg_write(_REG_PA_DAC, _PA_DAC_DEFAULT_VALUE)

        if "output_power" in lora_cfg:
            # See DS 3.4.2 RF Power Amplifiers
            dbm = int(lora_cfg["output_power"])
            if self._pa_boost:
                if dbm >= 20:
                    output_power = 0x15  # 17dBm setting
                    pa_dac = _PA_DAC_HIGH_POWER_20DBM
                else:
                    dbm = _clamp(dbm, 2, 17)  # +2 to +17dBm only
                    output_power = dbm - 2
                    pa_dac = _PA_DAC_DEFAULT_VALUE
                self._reg_write(_REG_PA_DAC, pa_dac)
            else:
                # In RFO mode, Output Power is computed from two register fields
                # - MaxPower and OutputPower.
                #
                # Do what the Semtech LoraMac-node driver does here, which is to
                # set max_power at one extreme or the other (0 or 7) and then
                # calculate the output_power setting based on this baseline.
                dbm = _clamp(dbm, -4, 15)
                if dbm > 0:
                    # MaxPower to maximum
                    pa_config |= _PA_CONFIG_MAXPOWER_MASK << _PA_CONFIG_MAXPOWER_SHIFT

                    # Pout (dBm) == 10.8dBm + 0.6*maxPower - (15 - register value)
                    # 10.8+0.6*7 == 15dBm, so pOut = register_value (0 to 15 dBm)
                    output_power = dbm
                else:
                    # MaxPower field will be set to 0

                    # Pout (dBm) == 10.8dBm - (15 - OutputPower)
                    # OutputPower == Pout (dBm) + 4.2
                    output_power = dbm + 4  # round down to 4.0, to keep using integer math

            pa_config |= output_power << _PA_CONFIG_OUTPUTPOWER_SHIFT
            update_mask |= (
                _PA_CONFIG_OUTPUTPOWER_MASK << _PA_CONFIG_OUTPUTPOWER_SHIFT
                | _PA_CONFIG_MAXPOWER_MASK << _PA_CONFIG_MAXPOWER_SHIFT
            )

        self._reg_update(_REG_PA_CONFIG, update_mask, pa_config)

        if "pa_ramp_us" in lora_cfg:
            # other fields in this register are reserved to 0 or unused
            self._reg_write(
                _REG_PA_RAMP,
                self._get_pa_ramp_val(
                    lora_cfg,
                    [10, 12, 15, 20, 25, 31, 40, 50, 62, 100, 125, 250, 500, 1000, 2000, 3400],
                ),
            )

        # If a hard reset happened then flags should be cleared already and mask should
        # default to fully enabled, but let's be "belts and braces" sure
        self._reg_write(_REG_IRQ_FLAGS, 0xFF)
        self._reg_write(_REG_IRQ_FLAGS_MASK, 0)  # do IRQ masking in software for now

        # Update MODEM_CONFIG1
        modem_config1 = 0x0
        update_mask = 0x0
        if "bw" in lora_cfg:
            bw = str(lora_cfg["bw"])
            bw_reg_val, self._bw_hz = {
                "7.8": (_MODEM_CONFIG1_BW7_8, 7800),
                "10.4": (_MODEM_CONFIG1_BW10_4, 10400),
                "15.6": (_MODEM_CONFIG1_BW15_6, 15600),
                "20.8": (_MODEM_CONFIG1_BW20_8, 20800),
                "31.25": (_MODEM_CONFIG1_BW31_25, 31250),
                "41.7": (_MODEM_CONFIG1_BW41_7, 41700),
                "62.5": (_MODEM_CONFIG1_BW62_5, 62500),
                "125": (_MODEM_CONFIG1_BW125, 125000),
                "250": (_MODEM_CONFIG1_BW250, 250000),
                "500": (_MODEM_CONFIG1_BW500, 500000),
            }[bw]
            modem_config1 |= bw_reg_val << _MODEM_CONFIG1_BW_SHIFT
            update_mask |= _MODEM_CONFIG1_BW_MASK << _MODEM_CONFIG1_BW_SHIFT

        if "freq_khz" in lora_cfg or "bw" in lora_cfg:
            # Workaround for Errata Note 2.1 "Sensitivity Optimization with a 500 kHz bandwidth"
            if self._bw_hz == 500000 and 862_000_000 <= self._rf_freq_hz <= 1020_000_000:
                self._reg_write(0x36, 0x02)
                self._reg_write(0x3A, 0x64)
            elif self._bw_hz == 500000 and 410_000_000 <= self._rf_freq_hz <= 525_000_000:
                self._reg_write(0x36, 0x02)
                self._reg_write(0x3A, 0x7F)
            else:
                # "For all other combinations of bandiwdth/frequencies, register at address 0x36
                # should be re-set to value 0x03 and the value at address 0x3a will be
                # automatically selected by the chip"
                self._reg_write(0x36, 0x03)

        if "coding_rate" in lora_cfg:
            self._coding_rate = int(lora_cfg["coding_rate"])
            if self._coding_rate < 5 or self._coding_rate > 8:
                raise ConfigError("coding_rate")
            # _MODEM_CONFIG1_CODING_RATE_45 == value 5 == 1
            modem_config1 |= (self._coding_rate - 4) << _MODEM_CONFIG1_CODING_RATE_SHIFT
            update_mask |= _MODEM_CONFIG1_CODING_RATE_MASK << _MODEM_CONFIG1_CODING_RATE_SHIFT

        if "implicit_header" in lora_cfg:
            self._implicit_header = lora_cfg["implicit_header"]
            modem_config1 |= _flag(_MODEM_CONFIG1_IMPLICIT_HEADER_MODE_ON, self._implicit_header)
            update_mask |= _MODEM_CONFIG1_IMPLICIT_HEADER_MODE_ON

        self._reg_update(_REG_MODEM_CONFIG1, update_mask, modem_config1)

        # Update MODEM_CONFIG2, for any fields that changed
        modem_config2 = 0
        update_mask = 0
        if "sf" in lora_cfg:
            sf = self._sf = int(lora_cfg["sf"])

            if sf < _MODEM_CONFIG2_SF_MIN or sf > _MODEM_CONFIG2_SF_MAX:
                raise ConfigError("sf")
            if sf == 6 and not self._implicit_header:
                # DS 4.1.12 "Spreading Factor"
                raise ConfigError("SF6 requires implicit_header mode")

            # Update these registers when writing 'SF'
            self._reg_write(
                _REG_DETECTION_THRESHOLD,
                _DETECTION_THRESHOLD_SF6 if sf == 6 else _DETECTION_THRESHOLD_OTHER,
            )
            # This field has a reserved non-zero field, so do a read-modify-write update
            self._reg_update(
                _REG_DETECT_OPTIMIZE,
                _DETECT_OPTIMIZE_AUTOMATIC_IF_ON | _DETECT_OPTIMIZE_MASK,
                _DETECT_OPTIMIZE_SF6 if sf == 6 else _DETECT_OPTIMIZE_OTHER,
            )

            modem_config2 |= sf << _MODEM_CONFIG2_SF_SHIFT
            update_mask |= _MODEM_CONFIG2_SF_MASK << _MODEM_CONFIG2_SF_SHIFT

        if "crc_en" in lora_cfg:
            self._crc_en = lora_cfg["crc_en"]
            # I had to double-check the datasheet about this point:
            # 1. In implicit header mode, this bit is used on both RX & TX and
            #    should be set to get CRC generation on TX and/or checking on RX.
            # 2. In explicit header mode, this bit is only used on TX (should CRC
            #    be added and CRC flag set in header) and ignored on RX (CRC flag
            #    read from header instead).
            modem_config2 |= _flag(_MODEM_CONFIG2_RX_PAYLOAD_CRC_ON, self._crc_en)
            update_mask |= _MODEM_CONFIG2_RX_PAYLOAD_CRC_ON

        self._reg_update(_REG_MODEM_CONFIG2, update_mask, modem_config2)

        # Update _REG_INVERT_IQ
        #
        # See comment about this register's undocumented weirdness at top of
        # file above _REG_INVERT_IQ constant.
        #
        # Note also there is a second register invert_iq2 which may be set differently
        # for transmit vs receive, see _set_invert_iq2() for that one.
        invert_iq = 0x0
        update_mask = 0x0
        if "invert_iq_rx" in lora_cfg:
            self._invert_iq[0] = lora_cfg["invert_iq_rx"]
            invert_iq |= _flag(_INVERT_IQ_RX, lora_cfg["invert_iq_rx"])
            update_mask |= _INVERT_IQ_RX
        if "invert_iq_tx" in lora_cfg:
            self._invert_iq[1] = lora_cfg["invert_iq_tx"]
            invert_iq |= _flag(_INVERT_IQ_TX_OFF, not lora_cfg["invert_iq_tx"])  # Inverted
            update_mask |= _INVERT_IQ_TX_OFF
        self._reg_update(_REG_INVERT_IQ, update_mask, invert_iq)

        if "preamble_len" in lora_cfg:
            self._preamble_len = lora_cfg["preamble_len"]
            self._reg_write(_REG_PREAMBLE_LEN_MSB, struct.pack(">H", self._preamble_len))

        # Update MODEM_CONFIG3, for any fields that have changed
        modem_config3 = 0
        update_mask = 0

        if "sf" in lora_cfg or "bw" in lora_cfg:
            # Changing either SF or BW means the Low Data Rate Optimization may need to be changed
            #
            # note: BaseModem.get_n_symbols_x4() assumes this value is set automatically
            # as follows.
            modem_config3 |= _flag(_MODEM_CONFIG3_LOW_DATA_RATE_OPTIMIZE, self._get_ldr_en())
            update_mask |= _MODEM_CONFIG3_LOW_DATA_RATE_OPTIMIZE

        if "lna_gain" in lora_cfg:
            lna_gain = lora_cfg["lna_gain"]
            update_mask |= _MODEM_CONFIG3_AGC_ON
            if lna_gain is None:  # Setting 'None' means 'Auto'
                modem_config3 |= _MODEM_CONFIG3_AGC_ON
            else:  # numeric register value
                # Clear the _MODEM_CONFIG3_AGC_ON bit, and write the manual LNA gain level 1-6
                # to the register
                self._reg_update(
                    _REG_LNA, _LNA_GAIN_MASK << _LNA_GAIN_SHIFT, lna_gain << _LNA_GAIN_SHIFT
                )

        if "rx_boost" in lora_cfg:
            self._reg_update(
                _REG_LNA,
                _LNA_BOOST_HF_MASK << _LNA_BOOST_HF_SHIFT,
                _flag(0x3, lora_cfg["lna_boost_hf"]),
            )

        self._reg_update(_REG_MODEM_CONFIG3, update_mask, modem_config3)

        if "syncword" in lora_cfg:
            self._reg_write(_REG_SYNC_WORD, lora_cfg["syncword"])

    def _reg_write(self, reg, value):
        self._cs(0)
        if isinstance(value, int):
            self._buf2[0] = reg | _WRITE_REG_BIT
            self._buf2[1] = value
            self._spi.write(self._buf2)
            if _DEBUG:
                dbg = hex(value)
        else:  # value is a buffer
            self._buf1[0] = reg | _WRITE_REG_BIT
            self._spi.write(self._buf1)
            self._spi.write(value)
            if _DEBUG:
                dbg = value.hex()
        self._cs(1)

        if _DEBUG:
            print("W {:#x} ==> {}".format(reg, dbg))
            self._reg_read(reg)  # log the readback as well

    def _reg_update(self, reg, update_mask, new_value):
        # Update register address 'reg' with byte value new_value, as masked by
        # bit mask update_mask. Bits not set in update_mask will be kept at
        # their pre-existing values in the register.
        #
        # If update_mask is zero, this function is a no-op and returns None.
        # If update_mask is not zero, this function updates 'reg' and returns
        # the previous complete value of 'reg' as a result.
        #
        # Note: this function has no way of detecting a race condition if the
        # modem updates any bits in 'reg' that are unset in update_mask, at the
        # same time a read/modify/write is occurring. Any such changes are
        # overwritten with the original values.

        if not update_mask:  # short-circuit if nothing to change
            if _DEBUG:
                # Log the current value if DEBUG is on
                # (Note the compiler will optimize this out otherwise)
                self._reg_read(reg)
            return
        old_value = self._reg_read(reg)
        value = ((old_value & ~update_mask) & 0xFF) | (new_value & update_mask)
        if old_value != value:
            self._reg_write(reg, value)
        return old_value

    def _reg_read(self, reg):
        # Read and return a single register value at address 'reg'
        self._buf2[0] = reg
        self._buf2[1] = 0xFF
        self._cs(0)
        self._spi.write_readinto(self._buf2, self._buf2)
        self._cs(1)
        if _DEBUG:
            print("R {:#x} <== {:#x}".format(reg, self._buf2[1]))
        return self._buf2[1]

    def _reg_readinto(self, reg, buf):
        # Read and return one or more register values starting at address 'reg',
        # into buffer 'buf'.
        self._cs(0)
        self._spi.readinto(self._buf1, reg)
        self._spi.readinto(buf)
        if _DEBUG:
            print("R {:#x} <== {}".format(reg, buf.hex()))
        self._cs(1)

    def _get_mode(self):
        # Return the current 'Mode' field in RegOpMode
        return self._reg_read(_REG_OPMODE) & _OPMODE_MODE_MASK

    def _set_mode(self, mode, lora_en=True):
        # Set the 'Mode' and 'LongRangeMode' fields in RegOpMode
        # according to 'mode' and 'lora_en', respectively.
        #
        # If enabling or disabling LoRa mode, the radio is automatically
        # switched into Sleep mode as required and then the requested mode is
        # set (if not sleep mode).
        #
        # Returns the previous value of the RegOpMode register (unmasked).
        mask = _OPMODE_LONGRANGEMODE_LORA | _OPMODE_MODE_MASK
        lora_val = _flag(_OPMODE_LONGRANGEMODE_LORA, lora_en)
        old_value = self._reg_read(_REG_OPMODE)
        new_value = (old_value & ~mask) | lora_val | mode

        if lora_val != (old_value & _OPMODE_LONGRANGEMODE_LORA):
            # Need to switch into Sleep mode in order to change LongRangeMode flag
            self._reg_write(_REG_OPMODE, _OPMODE_MODE_SLEEP | lora_val)

        if new_value != old_value:
            self._reg_write(_REG_OPMODE, new_value)

            if _DEBUG:
                print(
                    "Mode {} -> {} ({:#x})".format(
                        old_value & _OPMODE_MODE_MASK, mode, self._reg_read(_REG_OPMODE)
                    )
                )

        return old_value

    def _set_invert_iq2(self, val):
        # Set the InvertIQ2 register on/off as needed, unless it is already set to the correct
        # level
        if self._invert_iq[2] == val:
            return  # already set to the level we want
        self._reg_write(_REG_INVERT_IQ2, _INVERT_IQ2_ON if val else _INVERT_IQ2_OFF)
        self._invert_iq[2] = val

    def _standby(self):
        # Send the command for standby mode.
        #
        # **Don't call this function directly, call standby() instead.**
        #
        # (This private version doesn't update the driver's internal state.)
        old_mode = self._set_mode(_OPMODE_MODE_STDBY) & _OPMODE_MODE_MASK
        if old_mode not in (_OPMODE_MODE_STDBY, _OPMODE_MODE_SLEEP):
            # If we just cancelled sending or receiving, clear any pending IRQs
            self._reg_write(_REG_IRQ_FLAGS, 0xFF)

    def sleep(self):
        # Put the modem into sleep mode. Modem will wake automatically the next
        # time host asks it for something, or call standby() to wake it manually.
        self.standby()  # save some code size, this clears driver state for us
        self._set_mode(_OPMODE_MODE_SLEEP)

    def is_idle(self):
        # Returns True if the modem is idle (either in standby or in sleep).
        #
        # Note this function can return True in the case where the modem has temporarily gone to
        # standby, but there's a receive configured in software that will resume receiving the
        # next time poll_recv() or poll_send() is called.
        return self._get_mode() in (_OPMODE_MODE_STDBY, _OPMODE_MODE_SLEEP)

    def calibrate_image(self):
        # Run the modem Image & RSSI calibration process to improve receive performance.
        #
        # calibration will be run in the HF or LF band automatically, depending on the
        # current radio configuration.
        #
        # See DS 2.1.3.8 Image and RSSI Calibration. Idea to disable TX power
        # comes from Semtech's sx1276 driver which does this.

        pa_config = self._reg_update(_REG_PA_CONFIG, 0xFF, 0)  # disable TX power

        self._set_mode(_OPMODE_MODE_STDBY, False)  # Switch to FSK/OOK mode to expose RegImageCal

        self._reg_update(_REG_FSKOOK_IMAGE_CAL, _IMAGE_CAL_START, _IMAGE_CAL_START)
        while self._reg_read(_REG_FSKOOK_IMAGE_CAL) & _IMAGE_CAL_RUNNING:
            time.sleep_ms(1)

        self._set_mode(_OPMODE_MODE_STDBY)  # restore LoRA mode

        self._reg_write(_REG_PA_CONFIG, pa_config)  # restore previous TX power

    def calibrate(self):
        # Run a full calibration.
        #
        # For SX1276, this means just the image & RSSI calibration as no other runtime
        # calibration is implemented in the modem.
        self.calibrate_image()

    def start_recv(self, timeout_ms=None, continuous=False, rx_length=0xFF):
        # Start receiving.
        #
        # Part of common low-level modem API, see README.md for usage.
        super().start_recv(timeout_ms, continuous, rx_length)  # sets self._rx

        # will_irq if DIO0 and DIO1 both hooked up, or DIO0 and no timeout
        will_irq = self._dio0 and (self._dio1 or timeout_ms is None)

        if self._tx:
            # Send is in progress and has priority, _check_recv() will start receive
            # once send finishes (caller needs to call poll_send() for this to happen.)
            if _DEBUG:
                print("Delaying receive until send completes")
            return will_irq

        # Put the modem in a known state. It's possible a different
        # receive was in progress, this prevent anything changing while
        # we set up the new receive
        self._standby()  # calling private version to keep driver state as-is

        # Update the InvertIQ2 setting for RX
        self._set_invert_iq2(self._invert_iq[0])

        if self._implicit_header:
            # Payload length only needs to be set in implicit header mode
            self._reg_write(_REG_PAYLOAD_LEN, rx_length)

        if self._dio0:
            # Field value is 0, for DIO0 = RXDone
            update_mask = _DIO0_MAPPING_MASK << _DIO0_MAPPING_SHIFT
            if self._dio1:
                # Field value also 0, for DIO1 = RXTimeout
                update_mask |= _DIO1_MAPPING_MASK << _DIO1_MAPPING_SHIFT
            self._reg_update(_REG_DIO_MAPPING1, update_mask, 0)

        if not continuous:
            # Unlike SX1262, SX1276 doesn't have a "single RX no timeout" mode. So we set the
            # maximum hardware timeout and resume RX in software if needed.
            if timeout_ms is None:
                timeout_syms = 1023
            else:
                t_sym_us = self._get_t_sym_us()
                timeout_syms = (timeout_ms * 1000 + t_sym_us - 1) // t_sym_us  # round up

                # if the timeout is too long for the modem, the host will
                # automatically resume it in software. If the timeout is too
                # short for the modem, round it silently up to the minimum
                # timeout.
                timeout_syms = _clamp(timeout_syms, 4, 1023)
            self._reg_update(
                _REG_MODEM_CONFIG2,
                _MODEM_CONFIG2_SYMB_TIMEOUT_MSB_MASK,
                timeout_syms >> 8,
            )
            self._reg_write(_REG_SYMB_TIMEOUT_LSB, timeout_syms & 0xFF)

        # Allocate the full FIFO for RX
        self._reg_write(_REG_FIFO_ADDR_PTR, 0)
        self._reg_write(_REG_FIFO_RX_BASE_ADDR, 0)

        self._set_mode(_OPMODE_MODE_RX_CONTINUOUS if continuous else _OPMODE_MODE_RX_SINGLE)

        return will_irq

    def _rx_flags_success(self, flags):
        # Returns True if IRQ flags indicate successful receive.
        # Specifically, from the bits in _IRQ_DRIVER_RX_MASK:
        # - _IRQ_RX_DONE must be set
        # - _IRQ_RX_TIMEOUT must not be set
        # - _IRQ_PAYLOAD_CRC_ERROR must not be set
        # - _IRQ_VALID_HEADER must be set if we're using explicit packet mode, ignored otherwise
        return flags & _IRQ_DRIVER_RX_MASK == _IRQ_RX_DONE | _flag(
            _IRQ_VALID_HEADER, not self._implicit_header
        )

    def _get_irq(self):
        return self._reg_read(_REG_IRQ_FLAGS)

    def _clear_irq(self, to_clear=0xFF):
        return self._reg_write(_REG_IRQ_FLAGS, to_clear)

    def _read_packet(self, rx_packet, flags):
        # Private function to read received packet (RxPacket object) from the
        # modem, if there is one.
        #
        # Called from poll_recv() function, which has already checked the IRQ flags
        # and verified a valid receive happened.

        ticks_ms = self._get_last_irq()  # IRQ timestamp for the receive

        rx_payload_len = self._reg_read(_REG_RX_NB_BYTES)

        if rx_packet is None or len(rx_packet) != rx_payload_len:
            rx_packet = RxPacket(rx_payload_len)

        self._reg_readinto(_REG_FIFO, rx_packet)

        rx_packet.ticks_ms = ticks_ms
        # units: dB*4
        rx_packet.snr = self._reg_read(_REG_PKT_SNR_VAL)
        if rx_packet.snr & 0x80:  # Signed 8-bit integer
            # (avoiding using struct here to skip a heap allocation)
            rx_packet.snr -= 0x100
        # units: dBm
        rx_packet.rssi = self._reg_read(_REG_PKT_RSSI_VAL) - (157 if self._pa_boost else 164)
        rx_packet.crc_error = flags & _IRQ_PAYLOAD_CRC_ERROR != 0
        return rx_packet

    def prepare_send(self, packet):
        # Prepare modem to start sending. Should be followed by a call to start_send()
        #
        # Part of common low-level modem API, see README.md for usage.
        if len(packet) > 255:
            raise ValueError("packet too long")

        # Put the modem in a known state. Any current receive is suspended at this point,
        # but calling _check_recv() will resume it later.
        self._standby()  # calling private version to keep driver state as-is

        if self._ant_sw:
            self._ant_sw.tx(self._pa_boost)

        self._last_irq = None

        if self._dio0:
            self._reg_update(
                _REG_DIO_MAPPING1,
                _DIO0_MAPPING_MASK << _DIO0_MAPPING_SHIFT,
                1 << _DIO0_MAPPING_SHIFT,
            )  # DIO0 = TXDone

        # Update the InvertIQ2 setting for TX
        self._set_invert_iq2(self._invert_iq[1])

        # Allocate the full FIFO for TX
        self._reg_write(_REG_FIFO_ADDR_PTR, 0)
        self._reg_write(_REG_FIFO_TX_BASE_ADDR, 0)

        self._reg_write(_REG_PAYLOAD_LEN, len(packet))

        self._reg_write(_REG_FIFO, packet)

        # clear the TX Done flag in case a previous call left it set
        # (won't happen unless poll_send() was not called)
        self._reg_write(_REG_IRQ_FLAGS, _IRQ_TX_DONE)

    def start_send(self):
        # Actually start a send that was loaded by calling prepare_send().
        #
        # This is split into a separate function to allow more precise timing.
        #
        # The driver doesn't verify the caller has done the right thing here, the
        # modem will no doubt do something weird if prepare_send() was not called!
        #
        # Part of common low-level modem API, see README.md for usage.
        self._set_mode(_OPMODE_MODE_TX)

        self._tx = True

        return self._dio0 is not None  # will_irq if dio0 is set

    def _irq_flag_tx_done(self):
        return _IRQ_TX_DONE


# Define the actual modem classes that use the SyncModem & AsyncModem "mixin-like" classes
# to create sync and async variants.

try:
    from .sync_modem import SyncModem

    class SX1276(_SX127x, SyncModem):
        pass

    # Implementation note: Currently the classes SX1276, SX1277, SX1278 and
    # SX1279 are actually all SX1276. Perhaps in the future some subclasses with
    # software enforced limits can be added to this driver, but the differences
    # appear very minor:
    #
    # - SX1276 seems like "baseline" with max freq.
    # - SX1277 supports max SF level of 9.
    # - SX1278 supports max freq 525MHz, therefore has no RFO_HF and RFI_HF pins.
    # - SX1279 supports max freq 960MHz.
    #
    # There also appears to be no difference in silicon interface or register values to determine
    # which model is connected.
    SX1277 = SX1278 = SX1279 = SX1276

except ImportError:
    pass

try:
    from .async_modem import AsyncModem

    class AsyncSX1276(_SX127x, AsyncModem):
        pass

    # See comment above about currently identical implementations
    AsyncSX1277 = AsyncSX1278 = AsyncSX1279 = AsyncSX1276

except ImportError:
    pass



# Write characteristic from central to peripheral and time data rate.

from micropython import const
import time, machine, bluetooth

TIMEOUT_MS = 2000

_IRQ_CENTRAL_CONNECT = const(1)
_IRQ_CENTRAL_DISCONNECT = const(2)
_IRQ_GATTS_WRITE = const(3)
_IRQ_PERIPHERAL_CONNECT = const(7)
_IRQ_PERIPHERAL_DISCONNECT = const(8)
_IRQ_GATTC_CHARACTERISTIC_RESULT = const(11)
_IRQ_GATTC_CHARACTERISTIC_DONE = const(12)
_IRQ_GATTC_WRITE_DONE = const(17)
_IRQ_MTU_EXCHANGED = const(21)

# How long to run the test for.
_NUM_NOTIFICATIONS = const(40)
_MTU_SIZE = const(131)
_CHAR_SIZE = const(_MTU_SIZE - 3)

SERVICE_UUID = bluetooth.UUID("A5A5A5A5-FFFF-9999-1111-5A5A5A5A5A5A")
CHAR_UUID = bluetooth.UUID("00000000-1111-2222-3333-444444444444")
CHAR = (CHAR_UUID, bluetooth.FLAG_WRITE | bluetooth.FLAG_WRITE_NO_RESPONSE)
SERVICE = (SERVICE_UUID, (CHAR,))
SERVICES = (SERVICE,)

packet_sequence = 0
waiting_events = {}


def irq(event, data):
    if event == _IRQ_CENTRAL_CONNECT:
        waiting_events[event] = data[0]
    elif event == _IRQ_PERIPHERAL_CONNECT:
        waiting_events[event] = data[0]
    elif event == _IRQ_GATTS_WRITE:
        global packet_sequence
        conn_handle, attr_handle = data
        data = ble.gatts_read(attr_handle)
        if not (data[0] == packet_sequence and data[-1] == (256 - packet_sequence) & 0xFF):
            print("_IRQ_GATTS_WRITE data invalid:", packet_sequence, data)
        elif packet_sequence % 10 == 0:
            print("_IRQ_GATTS_WRITE", packet_sequence)
        packet_sequence += 1
    elif event == _IRQ_GATTC_CHARACTERISTIC_RESULT:
        # conn_handle, def_handle, value_handle, properties, uuid = data
        if data[-1] == CHAR_UUID:
            waiting_events[event] = data[2]
        else:
            return
    elif event == _IRQ_MTU_EXCHANGED:
        # ATT MTU exchange complete (either initiated by us or the remote device).
        conn_handle, mtu = data
        print("_IRQ_MTU_EXCHANGED:", mtu)

    if event not in waiting_events:
        waiting_events[event] = None


def wait_for_event(event, timeout_ms):
    t0 = time.ticks_ms()
    while time.ticks_diff(time.ticks_ms(), t0) < timeout_ms:
        if event in waiting_events:
            return waiting_events.pop(event)
        machine.idle()
    raise ValueError("Timeout waiting for {}".format(event))


# Acting in peripheral role.
def instance0():
    multitest.globals(BDADDR=ble.config("mac"))
    ((char_handle,),) = ble.gatts_register_services(SERVICES)
    ble.gatts_set_buffer(char_handle, _CHAR_SIZE)
    print("gap_advertise")
    ble.gap_advertise(20_000, b"\x02\x01\x06\x04\xffMPY")
    multitest.next()
    try:
        # Wait for central to connect to us.
        conn_handle = wait_for_event(_IRQ_CENTRAL_CONNECT, TIMEOUT_MS)
        # Wait for central to disconnect us.
        wait_for_event(_IRQ_CENTRAL_DISCONNECT, 30000)
        print("final packet_sequence:", packet_sequence)
    finally:
        ble.active(0)


# Acting in central role.
def instance1():
    global packet_sequence
    ((char_handle,),) = ble.gatts_register_services(SERVICES)
    multitest.next()
    try:
        # Connect to peripheral and then disconnect.
        print("gap_connect")
        ble.config(mtu=_MTU_SIZE)
        ble.gap_connect(*BDADDR)
        conn_handle = wait_for_event(_IRQ_PERIPHERAL_CONNECT, TIMEOUT_MS)
        ble.gattc_exchange_mtu(conn_handle)

        # Discover characteristics.
        ble.gattc_discover_characteristics(conn_handle, 1, 65535)
        value_handle = wait_for_event(_IRQ_GATTC_CHARACTERISTIC_RESULT, TIMEOUT_MS)
        wait_for_event(_IRQ_GATTC_CHARACTERISTIC_DONE, TIMEOUT_MS)

        # Send data!
        data = bytearray(ord("A") + (i % 64) for i in range(_CHAR_SIZE))
        for mode in (0, 1):
            ticks_start = time.ticks_ms()
            for i in range(_NUM_NOTIFICATIONS):
                data[0] = packet_sequence
                data[-1] = 256 - packet_sequence
                if packet_sequence % 10 == 0:
                    print("gattc_write", packet_sequence)
                if mode == 0:
                    while True:
                        try:
                            ble.gattc_write(conn_handle, value_handle, data, mode)
                            break
                        except OSError:
                            pass
                else:
                    ble.gattc_write(conn_handle, value_handle, data, mode)
                    wait_for_event(_IRQ_GATTC_WRITE_DONE, TIMEOUT_MS)
                packet_sequence += 1

            ticks_end = time.ticks_ms()
            ticks_total = time.ticks_diff(ticks_end, ticks_start)

            multitest.output_metric(
                "Did {} writes in {} ms. {} ms/write, {} bytes/sec".format(
                    _NUM_NOTIFICATIONS,
                    ticks_total,
                    ticks_total / _NUM_NOTIFICATIONS,
                    _NUM_NOTIFICATIONS * len(data) * 1000 // ticks_total,
                )
            )

            time.sleep_ms(100)

        # DIsconnect the peripheral.
        print("gap_disconnect:", ble.gap_disconnect(conn_handle))
        wait_for_event(_IRQ_PERIPHERAL_DISCONNECT, 20000)
    finally:
        ble.active(0)


ble = bluetooth.BLE()
ble.active(1)
ble.irq(irq)



# espnow module for MicroPython on ESP8266
# MIT license; Copyright (c) 2022 Glenn Moloney @glenn20

from _espnow import *
from select import poll, POLLIN


class ESPNow(ESPNowBase):
    # Static buffers for alloc free receipt of messages with ESPNow.irecv().
    _data = [bytearray(ADDR_LEN), bytearray(MAX_DATA_LEN)]
    _none_tuple = (None, None)

    def __init__(self):
        super().__init__()
        self._poll = poll()  # For any() method below...
        self._poll.register(self, POLLIN)

    def irecv(self, timeout_ms=None):
        n = self.recvinto(self._data, timeout_ms)
        return self._data if n else self._none_tuple

    def recv(self, timeout_ms=None):
        n = self.recvinto(self._data, timeout_ms)
        return [bytes(x) for x in self._data] if n else self._none_tuple

    def __iter__(self):
        return self

    def __next__(self):
        return self.irecv()  # Use alloc free irecv() method

    def any(self):  # For the ESP8266 which does not have ESPNow.any()
        try:
            next(self._poll.ipoll(0))
            return True
        except StopIteration:
            return False



import ffi
import sys


_h = None

names = ("libc.so", "libc.so.0", "libc.so.6", "libc.dylib")


def get():
    global _h
    if _h:
        return _h
    err = None
    for n in names:
        try:
            _h = ffi.open(n)
            return _h
        except OSError as e:
            err = e
    raise err


def set_names(n):
    global names
    names = n


# Find out bitness of the platform, even if long ints are not supported
# TODO: All bitness differences should be removed from micropython-lib, and
# this snippet too.
bitness = 1
v = sys.maxsize
while v:
    bitness += 1
    v >>= 1
